/**
 * Кэш данных.
 * @class Организация кэширования активно используемых данных. Включает набор функций для реализации механизма кэширования и работы с закэшированными данными.
 * @name y5.Cache
 *
 * @example
 * var cache = new y5.Cache();
 * cache.set("foo", "Bar");
 * cache.get("foo");
 * // -> "Bar"
 *
 * cache.test("foo");
 * // -> true
 *
 * cache.empty("baz");
 * // -> true
 *
 * cache.remove("foo");
 * cache.test("foo");
 * // -> false
 *
 * @todo добавить время жизни
 */
y5.Cache = function() {
    // объект для хранения данных
    this.data = {};
};

y5.Cache.prototype = {
    /**
     * Получает ранее сохраненные данные из кэша. Возвращает запрошенные данные, если они находятся в кэше. Если запрошенных данных в кэше нет, возвращает undefined.
     * @name y5.Cache.get
     * @memberOf y5.Cache
     * @function
     * @param {String} key Уникальный ключ
     * @returns {Object | undefined} Данные из кэша
     */
    get: function(key) {
        return this.data[key];
    },

    /**
     * Записывает данные в кэш. Возвращает данные, помещенные в кэш.
     * @name y5.Cache.set
     * @memberOf y5.Cache
     * @function
     * @param {String} key Уникальный ключ
     * @param {Object} value Объект
     * @returns {Object} Объект, помещенный в кэш
     */
    set: function(key, value) {
        return this.data[key] = value;
    },

    /**
     * Проверяет присутствие необходимых данных в кэше. Результатом проверки является true, если требуемые данные уже находятся в кэше.
     * @name y5.Cache.test
     * @memberOf y5.Cache
     * @function
     * @param {String} key Уникальный ключ
     * @returns {Boolean} Результат проверки
     */
    test: function(key) {
        return typeof this.data[key] != y5.UNDEF;
    },

    /**
     * Проверяет отсутствие необходимых данных в кэше. Результатом проверки является true, если требуемых данных в кэше нет.
     * @name y5.Cache.empty
     * @memberOf y5.Cache
     * @function
     * @param {String} key Уникальный ключ
     * @returns {Boolean} Результат проверки
     */
    empty: function(key) {
        return typeof this.data[key] == y5.UNDEF;
    },

    /**
     * Удаляет данные из кэша.
     * @name y5.Cache.remove
     * @memberOf y5.Cache
     * @function
     * @param {String} key Уникальный ключ
     */
    remove: function(key) {
        delete this.data[key];
    }
};

y5.loaded('Cache');