y5.require(['Cache', 'Strings'], function() {

var cacheRegexp = new y5.Cache(),
    cacheTest = new y5.Cache(),
    Strings = y5.Strings,

    /**
     * Разбивает строку по пробелам.
     * Если имя - RegExp, то возвращает массив, содержащий элемент.
     * @param {String | RegExp} names Имена классов
     * @returns {Array} Список классов
     * @private
     */
    split = function(names) {
        if (typeof names == 'string') {
            return names.split(' ');
        }

        // regexp
        if (typeof names.source != y5.UNDEF) {
            return [names];
        }

        return names;
    },

    /**
     * Проверяет переданный элемент и имя класса.
     * @param {String} method Метод
     * @param {HTMLElement} element Элемент для проверки
     * @param {String | RegExp} name Имя класса
     * @private
     */
    checkElem = function(method, elem, name) {
        if (typeof elem == y5.UNDEF || elem == null) {
            throw new y5.Exception('object required', method, 'Classes');
        }
        if (!name || (typeof name != 'string' && !name.source)) {
            throw new y5.Exception('class name required', method, 'Classes');
        }
    },

    /**
     * Возвращает RegExp для работы с классами.
     * @param {String | RegExp} name Имя класса
     * @returns {RegExp} Правило для нахождения класса
     * @private
     */
    rName = function(name, nocache) {
        var _name = '',
            flags = '',
            source = '';

        if (typeof name == 'string') {
            source = _name = Strings.escapeRegexp(name);
        } else {
            source = name.source;
            _name = name.toString();
            // остальные флаги не нужны, т.к. не имеют смысла для поиска класса
            flags += name.ignoreCase?'i':'';
        }

        if (!nocache) {
            if (cacheRegexp.empty(_name)) {
                return cacheRegexp.set(_name, new RegExp('(^|\\s+)' + source + '(\\s+|$)', flags));
            }
            return cacheRegexp.get(_name);
        }
        return new RegExp('(^|\\s+)' + source + '(\\s+|$)', flags);
    };

/**
 * @class Набор функций для работы с классами CSS.
 * Включает реализацию функций, упрощающих работу с CSS-классами HTML-элементов страницы.
 * @static
 * @name y5.Classes
 * @memberOf y5
 */
y5.Classes = {
    /**
     * Проверяет элемент на наличие у него заданного CSS-класса. Имя класса может быть задано строкой или регулярным выражением. Метод возвращает true, если HTML-элемент имеет требуемый класс.
     * @name y5.Classes.test
     * @memberOf y5.Classes
     * @function
     * @param {HTMLElement} element Элемент
     * @param {String | RegExp} name Имя класса
     * @param {Boolean} [nocache] Не кэшировать результат проверки
     * @returns {Boolean} Результат проверки
     *
     * @example
     * // &lt;div class="one two"/&gt;
     * y5.Classes.test(element, 'two'); // -> true
     */
    test: function(elem, name, nocache) {
        checkElem('test', elem, name);

        if (name == '*') {
            return true;
        }
        try {
            if (!nocache) {
                var n = name + ' ' + elem.className;
                if (cacheTest.empty(n)) {
                    return cacheTest.set(n, rName(name).test(elem.className));
                }
                return cacheTest.get(n);
            }
            return rName(name, nocache).test(elem.className);
        } catch (e) {}

        return false;
    },

    /**
     * Присваивает элементу новое имя CSS-класса вместо существующего. Метод возвращает true, если класс был успешно установлен, или false, если HTML-элемент страницы уже имеет класс с данным именем.
     * @name y5.Classes.set
     * @memberOf y5.Classes
     * @function
     * @param {HTMLElement} element Элемент
     * @param {String} name Имя класса
     * @returns {Boolean} true, если класс был установлен
     *
     * @example
     * // &lt;div class="one"/>
     * y5.Classes.set(element, 'two'); // -> true
     * // &lt;div class="two"/>
     *
     * // &lt;div class="one"/>
     * y5.Classes.set(element, 'one'); // -> false
     * // &lt;div class="one"/>
     */
    set: function(elem, name) {
        checkElem('set', elem, name);
        if (elem.className != name) {
            elem.className = name;
            return true;
        }
        return false;
    },

    /**
     * Добавляет новые CSS-классы заданному элементу. В зависимости от типа передаваемых в параметре name данных, можно добавлять один или несколько классов одновременно. Если в качестве параметра name передается строка, то к существующим классам будет добавлен один новый класс с именем, соответствующим данной строке. Если в качестве параметра name передается массив, то к существующим классам будет добавлено несколько новых классов с именами, соответствующими элементам массива. Метод возвращает массив, содержащий названия добавленных классов.
     * @name y5.Classes.add
     * @memberOf y5.Classes
     * @function
     * @param {HTMLElement} element Элемент
     * @param {String | Array} name Имя класса
     * @returns {Array} Добавленные классы
     *
     * @example
     * // &lt;div class="one"/>
     * y5.Classes.add(element, 'two'); // -> ['two']
     * // &lt;div class="one two"/>
     *
     * // &lt;div class="one"/>
     * y5.Classes.add(element, ['foo', 'bar']); // -> ['foo', 'bar']
     * // &lt;div class="one foo bar"/>
     *
     * // &lt;div class="one"/>
     * y5.Classes.add(element, 'one foo bar'); // -> ['one', 'foo', 'bar']
     * // &lt;div class="one foo bar"/>
     */
    add: function(elem, names) {
        var addons = split(names).filter(
            function(name) {
                return !this.test(elem, name)
            },
            this
        );

        if (addons.length) {
            elem.className += ' ' + addons.join(' ');
        }

        return addons;
    },

    /**
     * Удаляет CSS-классы у заданного элемента. В зависимости от типа передаваемых в параметре name данных, можно удалять один или несколько классов одновременно. Если в качестве параметра name передается строка, то удаляется один класс с именем, соответствующим переданной строке. Если в качестве параметра name передается массив, то будет удалено несколько классов с именами, соответствующими элементам массива. Метод возвращает массив, содержащий названия удаленных классов.
     * @name y5.Classes.remove
     * @memberOf y5.Classes
     * @function
     * @param {HTMLElement} element Элемент
     * @param {String | Array} name Имя класса
     * @returns {Array} Удаленные классы
     *
     * @example
     * // &lt;div class="one two"/>
     * y5.Classes.remove(element, 'two'); // -> ['two']
     * // &lt;div class="one"/>
     */
    remove: function(elem, names) {
        var className = elem.className;
        var removed = [];

        split(names).forEach(
            function(name) {
                while (rName(name).test(className)) {
                    removed.push(name);
                    className = className.replace(rName(name), ' ');
                }
            }
        );

        elem.className = Strings.normalizeSpace(className);

        return removed;
    },

    /**
     * Заменяет один CSS-класс на другой у заданного элемента. Имя заменяемого класса может быть задано строкой или регулярным выражением. Метод возвращает true, если класс был успешно заменен.
     * @name y5.Classes.replace
     * @memberOf y5.Classes
     * @function
     * @param {HTMLElement} element Элемент
     * @param {String | RegExp} find Заменяемое имя класса
     * @param {String} replace Новое имя класса
     * @returns {Boolean} true, если класс find был заменен на replace
     *
     * @example
     * // &lt;div class="one two"/>
     * y5.Classes.replace(element, 'two', 'three'); // -> true
     * // &lt;div class="one three"/>
     */
    replace: function (elem, find, replace) {
        if (this.test(elem, find)) {
            elem.className = Strings.normalizeSpace(elem.className.replace(rName(find, true), '$1' + replace + '$2'));
            return true;
        }
        return false;
    },

    /**
     * Добавляет/удаляет CSS-класс элемента в зависимости от условия. Если значение параметра test равно true, то класс с именем, переданным в name, будет добавлен. Если значение параметра test рано false, то класс с именем, переданным в name, будет удален. Данный метод поддерживает операции добавления/удаления одновременно нескольких классов. Для этого в параметре name необходимо передавать массив имен классов. В качестве результата возвращается массив названий добавленных/удаленных классов.
     * @name y5.Classes.assign
     * @memberOf y5.Classes
     * @function
     * @param {HTMLElement} element Элемент
     * @param {String | Array} name Имя класса
     * @param {Boolean} test Если значение равно true, то добавляется новый класс, иначе - удаляется
     * @returns {Array} Добавленные или удаленные классы
     *
     * @example
     * // &lt;div class="one"/>
     * y5.Classes.assign(element, 'two', true); // -> ['two']
     * // &lt;div class="one two"/>
     *
     * // &lt;div class="one two"/>
     * y5.Classes.assign(element, 'two', false); // -> ['two']
     * // &lt;div class="one"/>
     */
    assign: function (elem, name, test) {
        if (test) {
            return this.add(elem, name);
        } else {
            return this.remove(elem, name);
        }
    },

    /**
     * Устанавливает/сбрасывает CSS-класс у заданного элемента. Класс будет сброшен (удален), если HTML-элемент уже имеет класс с заданным именем. Класс будет установлен (добавлен), если у HTML-элемента нет класса с заданным именем. Метод возвращает true в случае установки CSS-класса или false в случае его сброса.
     * @name y5.Classes.toggle
     * @memberOf y5.Classes
     * @function
     * @param {HTMLElement} element Элемент
     * @param {String} name Имя класса
     * @returns {Boolean} true, если класс был добавлен
     *
     * @example
     * // &lt;div class="one two"/>
     * y5.Classes.toggle(element, 'two'); // -> false
     * // &lt;div class="one"/>
     *
     * // &lt;div class="one"/>
     * y5.Classes.toggle(element, 'two'); // -> true
     * // &lt;div class="one two"/>
     */
    toggle: function (elem, name) {
        var added = !this.test(elem, name);
        this.assign(elem, name, added);
        return added;
    },

    /**
     * Переключает CSS-класс у заданного HTML-элемента страницы с одного значения на другое. В качестве параметров передаются два имени классов, и HTML-элементу присваивается тот, который отличается от текущего. Метод возвращает имя CSS-класса, на который был переключен исходный класс.
     * @name y5.Classes.swap
     * @memberOf y5.Classes
     * @function
     * @param {HTMLElement} element Элемент
     * @param {String} name1 Имя класса
     * @param {String} name2 Имя класса
     * @returns {String} Класс, который заменил собой искомый
     *
     * @example
     * // &lt;div class="one"/>
     * y5.Classes.swap(element, 'two', 'one'); // -> 'two'
     * // &lt;div class="two"/>
     *
     * // &lt;div class="two"/>
     * y5.Classes.swap(element, 'two', 'one'); // -> 'one'
     * // &lt;div class="one"/>
     */
    swap: function (elem, one, two) {
        if (this.test(elem, one)) {
            this.replace(elem, one, two);
            return two;
        } else if (this.test(elem, two)) {
            this.replace(elem, two, one);
        } else {
            this.add(elem, one);
        }
        return one;
    }
};

y5.loaded('Classes');

});
