/**
 * Ядро фреймворка.
 * Содержит определения часто используемых переменных (версия браузера, тип операционной системы и т.п.) и набор функций для загрузки модулей. 
 * 
 * @class
 * @static
 * @name y5
 *
 * @example
 * Для загрузки модулей библиотек можно использовать удобный механизм динамической загрузки.
 * Для этого используется метод y5.require.
 *
 * y5.require('Dom', function() {
 *     // здесь можно использовать функции y5.Dom
 * });
 *
 * y5.require('URL', 'Classes', function() {
 *     // здесь можно использовать функции y5.URL и y5.Classes
 * });
 */
var y5 = (function() {

    var
        // Список загруженных модулей
        // y5:Arrays, y5:Dom, ...
        Modules = {},
        
        /**
         * Массив заблокированных модулей
         * @name BlockModules
         * @memberOf y5
         * @type Array
         * @private
         */
        BlockModules = [],

        // Список требуемых модулей
        ModulesRequired = {},

        // Правило для вычисления названия библиотеки и имени модуля
        // y5:Arrays или {y5}.Arrays (устарело)
        ModuleRegexp = /^(\{([^\}]+)\}\.|([^:]+):)?(.+)$/,

        // Список библиотек
        // y5, Friends, ...
        // Формат: {y5:
        //   {
        //     path: [путь к библиотеке],
        //     charset: [кодировка файлов модулей],
        //     query: [параметры URL, добавляемые к запросу модуля]
        //   }
        // }
        Namespaces = {},

        // Имя библиотеки по умолчанию
        NamespaceDefault = 'y5',

        // Очередь модулей
        // [y5:Arrays, y5:Dom, ...]
        Query = [],

        // Список скриптов страницы
        Scripts = document.getElementsByTagName('script'),

        moduleLoadedNotifier,
        moduleRequiredNotifier,
        namespaceAddedNotifier;

    function runCallback(callback) {
        if (typeof callback == 'function') {
            callback();
        }
    }

    function isModuleLoaded(module) {
        return !!Modules[module];
    }

    function isModuleNotLoaded(module) {
        return !Modules[module];
    }

    function isModulesLoaded(modules) {
        return modules.every(isModuleLoaded);
    }
    
    function isModuleBlocked (module) {
        for (var i=0,j=BlockModules.length; i<j; i++) {
            if (module.indexOf(BlockModules[i]) == 0) {
                return true;
            }
        }
        return false;
    }

    function addModuleLoaded(module) {
        Modules[module] = 1;

        if (!moduleLoadedNotifier) {
            moduleLoadedNotifier = y5.Events.notify('y5:moduleLoaded', y5, false);
        }
        moduleLoadedNotifier.dispatch(module);
    }

    function isModuleRequired(module) {
        return !!ModulesRequired[module];
    }

    function addModuleRequired(module) {
        ModulesRequired[module] = 1;

        if (!moduleRequiredNotifier) {
            moduleRequiredNotifier = y5.Events.notify('y5:moduleRequired', y5, false);
        }
        moduleRequiredNotifier.dispatch(module);
    }

    function removeModuleRequired(module) {
        delete ModulesRequired[module];
    }

    function addNamespace(namespace, data) {
        Namespaces[namespace] = data;
        if (!namespaceAddedNotifier) {
            namespaceAddedNotifier = y5.Events.notify('y5:namespaceAdded', y5, false);
        }
        namespaceAddedNotifier.dispatch(data);
        callQuery();
    }

    function getNamespaceData(namespace) {
        return Namespaces[namespace];
    }

    function getModuleData(module) {
        var data = module.match(ModuleRegexp);
        // TODO: exception
        return [
            data[3] || data[2] || NamespaceDefault,
            data[4]
        ];
    }

    function getModuleNameCanonical(module) {
        var data = getModuleData(module);
        return data[0] + ':' + data[1];
    }

    function getModuleURL(module, type) {
        var data = getModuleData(module);
        var namespace = getNamespaceData(data[0]);

        if (!namespace) {
            return false;
        }

        return [
            (
                namespace.path +
                data[1].replace(/\./g, '/') +
                '.' + (type || 'js') +
                namespace.query
            ),
            namespace.charset
        ];
    }

    /**
     * Загружает модули по необходимости.
     * @private
     */
    function loadModules(modules) {
        var i = 0,
            l = modules.length,
            module;
        for (; i < l; i++) {
            module = modules[i];
            if (isModuleLoaded(module) || isModuleRequired(module) || isModuleBlocked(module)) {
                continue;
            }

            var data = getModuleURL(module);
            if (data) {
                addModuleRequired(module);
                y5.Loader.loadScript(data[0], data[1]);
            }
        }
    }

    /**
     * Помещает в очередь список требуемых модулей и callback-функцию.
     * @private
     */
    function pushQuery(modules, callback) {
        modules = modules.filter(isModuleNotLoaded);

        if (modules.length == 0) {
            runCallback(callback);
            return false;
        }

        Query.push({
            modules: modules,
            callback: callback
        });

        return modules;
    }

    /**
     * Запускает callback-функции из очереди для загруженных модулей.
     * @private
     */
    function callQuery() {
        for (var i = 0; i < Query.length; i++) {
            var item = Query[i];
            if (isModulesLoaded(item.modules)) {
                Query.splice(i, 1);
                runCallback(item.callback);
                i--;
            } else {
                loadModules(item.modules);
            }
        }
    }

    /**
     * Возвращает данные скрипта по имени файла.
     * @param {String} file Имя файла
     * @param {String} [charset] Кодировка файла (если указано, то переопределяет кодировку указанную в скрипте)
     * @returns {Object} Объект вида {path: 'http://www.yandex.ru/foo/', query: '?build=1', charset: 'utf-8'}
     * @private
     */
    function getData(file, charset) {
        for (var i = 0, l = Scripts.length; i < l; i++) {
            var script = Scripts[i],
                src = script.getAttribute('src');

            if (src && src.lastIndexOf(file) >= 0) {
                var base = {
                    path: src.substring(0, src.lastIndexOf('/') + 1),
                    charset: charset || script.getAttribute('charset') || 'utf-8'
                };

                var query = src.lastIndexOf('?');
                if (query >= 0) {
                    base.query = src.substring(query, src.length);
                }

                return base;
            }
        }
        return null;
    }

    function registerNamespaceByData(namespace, data, charset) {
        if (typeof data == 'string') {
            data = {path: data, charset: charset};
        }

        // дополняем путь слэшем, если забыт
        if (data.path.lastIndexOf('/') != data.path.length - 1) {
            data.path += '/';
        }

        // устанавливаем кодировку по умолчанию
        if (!data.charset) {
            data.charset = 'utf-8';
        }

        // устанавливаем параметры запроса по умолчанию
        if (!data.query) {
            data.query = '';
        }

        addNamespace(namespace, data);
    }

    function getNamespaceItem(namespace, name) {
        try {
            return Namespaces[namespace][name];
        } catch (e) {
            return null;
        }
    }

    var obj = {
        /**
         * Текущая версия фреймворка.
         * @name y5.version
         * @memberOf y5
         * @type Number
         */
        version: 1.5,

        /**
         * Логическая переменная. Устанавливается в true, когда происходит событие dom:loaded.
         * @name y5.domloaded
         * @memberOf y5
         * @type Boolean
         */
        domloaded: false,

        /**
         * Загружает требуемые модули и выполняет callback-функцию после загрузки модулей.
         * @name y5.require
         * @memberOf y5
         * @function
         * @param {String | Array} modules "foo.bar"
         * @param {Function} callback Функция, выполняемая в момент загрузки
         */
        require: function(/* module1, module2, ..., moduleN [ , callback ] */) {
            var args = arguments,
                modules = [],
                callback = y5.VOID,
                Types = y5.Types;

            for (var i = 0, l = args.length; i < l; i++) {
                var arg = args[i];
                switch (Types.type(arg)) {
                    case Types.ARRAY:
                        modules = modules.concat(arg);
                        break;

                    case Types.STRING:
                        modules.push(arg);
                        break;

                    case Types.FUNCTION:
                        callback = arg;
                        break;
                }
            }

            modules = modules.map(getModuleNameCanonical);
            modules = pushQuery(modules, callback);

            if (modules) {
                loadModules(modules);
            }
        },

        /**
         * Функция вызывается модулем после окончания загрузки всех зависимых модулей.
         * @name y5.loaded
         * @memberOf y5
         * @function
         * @param {String} module Имя модуля (например, "foo.bar")
         */
        loaded: function(module) {
            module = getModuleNameCanonical(module);
            removeModuleRequired(module);
            addModuleLoaded(module);
            callQuery();
        },

        /**
         * Регистрирует новую библиотеку модулей по имени файла.
         * @name y5.registerNamespace
         * @memberOf y5
         * @function
         * @param {String} namespace Имя, под которым регистрируется библиотека
         * @param {String} filename Строка, которая входит в имя файла библиотеки
         * @param {String} [charset] Кодировка файлов библиотеки
         */
        registerNamespace: function(namespace, file, charset) {
            var timer = null,
                counter = 0;

            function get() {
                if (counter < 1000) {
                    var data = getData(file, charset);
                    if (data) {
                        window.clearTimeout(timer);
                        registerNamespaceByData(namespace, data);
                        return true;
                    }
                    counter++;
                }
                return false;
            }

            if (!get()) {
                timer = window.setInterval(get, 1);
            }
        },

        /**
         * Регистрирует новую библиотеку модулей по URL или объекту с данными о библиотеке.
         * @name y5.registerNamespaceByData
         * @memberOf y5
         * @function
         * @param {String} namespace Имя, под которым регистрируется библиотека
         * @param {Object | String} data Данные вида: {path: 'http://www.yandex.ru/foo/', query: '?build=1', charset: 'utf-8'}
         * или путь к библиотеке.
         * @param {String} [charset] Кодировка файлов библиотеки
         */
        registerNamespaceByData: registerNamespaceByData,

        /**
         * Возвращает путь к файлам библиотеки.
         * @name y5.namespacePath
         * @memberOf y5
         * @function
         * @param {String} namespace Имя библиотеки
         * @returns {String | null} Путь
         *
         * @example
         * y5.namespacePath("Foo");
         * // -> "http://www.yandex.ru/foo/"
         */
        namespacePath: function(namespace) {
            return getNamespaceItem(namespace, 'path');
        },

        /**
         * Возвращает кодировку файлов библиотеки.
         * @name y5.namespaceCharset
         * @memberOf y5
         * @function
         * @param {String} namespace Имя библиотеки
         * @returns {String | null} Кодировка
         *
         * @example
         * y5.namespaceCharset("Foo");
         * // -> "utf-8"
         */
        namespaceCharset: function(namespace) {
            return getNamespaceItem(namespace, 'charset');
        },

        /**
         * Возвращает URL модуля.
         * @name y5.moduleURL
         * @memberOf y5
         * @function
         * @param {String} module Имя модуля (например, "foo.bar")
         * @param {String} ext Расширение имени файла модуля (по умолчанию - "js")
         * @returns {String} URL модуля
         *
         * @example
         * y5.moduleURL("Dom");
         * // -> [base URL]/Dom.js
         *
         * y5.moduleURL("Widget.Window", "css");
         * // -> [base URL]/Widget/Window.css
         */
        moduleURL: function(module, ext) {
            return getModuleURL(module, ext)[0];
        },

        /**
         * Возвращает каноническое (полное) имя модуля.
         * @name y5.moduleName
         * @memberOf y5
         * @function
         * @param {String} module Имя модуля (например, "foo.bar")
         * @returns {String} Имя модуля
         *
         * @example
         * y5.moduleName("foo.bar");
         * // -> "y5:foo.bar"
         */
        moduleName: getModuleNameCanonical,

        /**
         * Возвращает название библиотеки модуля.
         * @name y5.moduleNamespace
         * @memberOf y5
         * @function
         * @param {String} module Имя модуля (например, "foo.bar")
         * @returns {String} Namespace модуля
         *
         * @example
         * y5.moduleNamespace("foo.bar");
         * // -> "y5"
         */
        moduleNamespace: function(module) {
            return getModuleData(module)[0];
        },

        /**
         * Возвращает объект по имени модуля.
         * @name y5.moduleObject
         * @memberOf y5
         * @function
         * @param {String} module Имя модуля (например, "foo.bar")
         * @returns {Object | null} Объект
         *
         * @example
         * y5.moduleNamespace("foo.bar");
         * // -> y5.foo.bar
         */
        moduleObject: function(module) {
            var obj = window;
            var classes = getModuleNameCanonical(module).split(/[:\.]/g);

            for (var i = 0, l = classes.length; i < l; i++) {
                obj = obj[classes[i]];
                if (typeof obj == y5.UNDEF) {
                    return null;
                }
            }

            return obj;
        },
        
        /**
         * Блокирует загрузку дочерних модулей для указанного модуля или библиотеки.
         * @name y5.blockLoad
         * @memberOf y5
         * @function 
         * @param {String} module Имя модуля (например, "foo.bar") или библиотеки
         * @param {Boolean} withModule Блокировать также загрузку указанного модуля
         * @returns  {Boolean}
         */
        blockLoad: function (module, withModule) {
            // не неймспейс
            if (module.indexOf(':') != (module.length-1)) {
                module = getModuleNameCanonical(module);
                // если блокируем только дочерние модули, то добавляем в конец имени точку
                if (!withModule) {
                    module += '.';
                }
            }
            
            // если модуля нет в списке 
            if (module && BlockModules.indexOf(module) == -1) {
                BlockModules.push(module);
                return true;
            }
            return false;
        },
        
        /**
         * Разблокирует загрузку дочерних модулей для указанного модуля или библиотеки.
         * @name y5.unblockLoad
         * @memberOf y5
         * @function 
         * @param {String} module Имя модуля (например, "foo.bar") или библиотеки
         * @returns {Boolean}
         */
        unblockLoad: function (module) {
            // не неймспейс
            if (module.indexOf(':') != (module.length-1)) {
                module = getModuleNameCanonical(module);
            }
            
            if (module) {
                var index = BlockModules.indexOf(module);
                if (index == -1) {
                    index = BlockModules.indexOf(module + '.');
                }
                if (index != -1) {
                    BlockModules.splice(index, 1);
                    return true;
                }
            }
            return false;
        }
    };

    /**
     * Возвращает данные скрипта по имени файла.
     * @param {String} file Имя файла
     * @param {String} [charset] Кодировка файла (если указана, то переопределяет кодировку, указанную в скрипте)
     * @returns {Object} Объект вида {path: 'http://www.yandex.ru/foo/', query: '?build=1', charset: 'utf-8'}
     * @name getBase
     * @memberOf y5
     * @function
     * @deprecated
     */
    obj.getBase = getData;
    obj.getBaseAndSetAlias = obj.registerNamespace;
    obj.setAlias = registerNamespaceByData;
    obj.constructURL = obj.moduleURL;
    obj.getAlias = obj.moduleNamespace;
    obj.charsets = {};
    obj.setAliasCharset = function() {};

    return obj;

})();

/// Vars
/////////////////////////////////////////////////////////////////////

/**
 * @class Содержит глобальные переменные.
 * @static
 * @name y5.Vars
 */
y5.Vars = {
    DEBUG: false,

    /**
     * Строка 'undefined'.
     * @name y5.Vars.UNDEF
     * @memberOf y5.Vars
     * @constant
     * @type String
     */

    /**
     * @name y5.UNDEF
     * @link y5.Vars.UNDEF
     * @memberOf y5
     * @type String
     */
    UNDEF: 'undefined',

    /**
     * Всегда возвращает значение false.
     * @name y5.Vars.FALSE
     * @memberOf y5.Vars
     * @function
     * @type Function
     * @returns {Boolean} false
     */

    /**
     * @name y5.FALSE
     * @link y5.Vars.FALSE
     * @memberOf y5
     * @type Function
     */
    FALSE: function() { return false; },

    /**
     * Всегда возвращает значение true.
     * @name y5.Vars.TRUE
     * @memberOf y5.Vars
     * @function
     * @type Function
     * @returns {Boolean} true
     */

    /**
     * @name y5.TRUE
     * @link y5.Vars.TRUE
     * @memberOf y5
     * @type Function
     */
    TRUE: function() { return true; },

    /**
     * Всегда возвращает значение null.
     * @name y5.Vars.NULL
     * @memberOf y5.Vars
     * @function
     * @type Function
     * @returns {Object} null
     */

    /**
     * @name y5.NULL
     * @link y5.Vars.NULL
     * @memberOf y5
     * @type Function
     */
    NULL: function() { return null; },

    /**
     * Всегда возвращает значение undefined.
     * @name y5.Vars.VOID
     * @memberOf y5.Vars
     * @function
     * @type Function
     * @returns {undefined}
     */

    /**
     * @name y5.VOID
     * @link y5.Vars.VOID
     * @memberOf y5
     * @type Function
     */
    VOID: function() { }

};

/// Browser
/////////////////////////////////////////////////////////////////////

y5.Browser = {
    get: function(nav) {
        var agt = nav.userAgent.toLowerCase(),
            result = {};

        function getVer(is_this, regex) {
            if (is_this) {
                var f = agt.match(regex);
                return f ? parseFloat(f[1]) : 0;
            }
            return 0;
        }

        function contains(needle) {
            return agt.indexOf(needle) != -1;
        }

        /**
         * Определяет операционную систему пользователя. Возвращает true, если операционная система пользователя - MS Windows.
         * @name y5.Vars.is_win
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_win
         * @link y5.Vars.is_win
         * @memberOf y5
         * @type Boolean
         */
        result.is_win       = contains('windows');

        /**
         * Определяет операционную систему пользователя. Возвращает true, если операционная система пользователя - Mac OS.
         * @name y5.Vars.is_mac
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_mac
         * @link y5.Vars.is_mac
         * @memberOf y5
         * @type Boolean
         */
        result.is_mac       = contains('mac');
        
        /**
         * Определяет операционную систему пользователя. Возвращает true, если операционная система пользователя - GNU/Linux.
         * @name y5.Vars.is_linux
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_linux
         * @link y5.Vars.is_linux
         * @memberOf y5
         * @type Boolean
         */
        result.is_linux       = contains('linux');

        /**
         * Определяет тип браузера. Возвращает true для браузера Safari.
         * @name y5.Vars.is_safari
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_safari
         * @link y5.Vars.is_safari
         * @memberOf y5
         * @type Boolean
         */
        result.is_safari    = contains('safari');

        /**
         * Определяет тип браузера. Возвращает true для браузера Safari на iPhone и iPod Touch.
         * @name y5.Vars.is_iphone
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_iphone
         * @link y5.Vars.is_iphone
         * @memberOf y5
         * @type Boolean
         */
        result.is_iphone    = result.is_safari && contains('iphone');

        /**
         * Определяет тип браузера. Возвращает true для браузера Opera.
         * @name y5.Vars.is_opera
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_opera
         * @link y5.Vars.is_opera
         * @memberOf y5
         * @type Boolean
         */
        result.is_opera     = contains('opera');

        /**
         * Определяет тип браузера. Возвращает true для браузера Konqueror.
         * @name y5.Vars.is_konq
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_konq
         * @link y5.Vars.is_konq
         * @memberOf y5
         * @type Boolean
         */
        result.is_konq      = contains('konqueror');

        /**
         * Определяет тип браузера. Возвращает true для браузеров семейства MS Internet Explorer.
         * @name y5.Vars.is_ie
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie
         * @link y5.Vars.is_ie
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie        = !result.is_opera && contains('msie');

        /**
         * Определяет тип движка браузера. Возвращает true для браузеров, основанных на KHTML.
         * @name y5.Vars.is_khtml
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_khtml
         * @link y5.Vars.is_khtml
         * @memberOf y5
         * @type Boolean
         */
        result.is_khtml     = !result.is_safari && contains('khtml');

        /**
         * Определяет тип движка браузера. Возвращает true, если в браузере используется движок Gecko.
         * @name y5.Vars.is_gecko
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_gecko
         * @link y5.Vars.is_gecko
         * @memberOf y5
         * @type Boolean
         */
        result.is_gecko     = contains('gecko/');

        /**
         * Определяет версию браузера семейства MS Internet Explorer. Возвращает номер версии.
         * @name y5.Vars.ie_ver
         * @memberOf y5.Vars
         * @constant
         * @type Number
         */

        /**
         * @name y5.ie_ver
         * @link y5.Vars.ie_ver
         * @memberOf y5
         * @type Number
         */
        result.ie_ver       = getVer(result.is_ie, /msie (\d+\.\d)/);

        /**
         * Определяет версию движка рендеринга страниц для браузеров, основанных на коде проекта Gecko. Возвращает номер версии.
         * @name y5.Vars.gecko_ver
         * @memberOf y5.Vars
         * @constant
         * @type Number
         */

        /**
         * @name y5.gecko_ver
         * @link y5.Vars.gecko_ver
         * @memberOf y5
         * @type Number
         */
        result.gecko_ver    = getVer(result.is_gecko, /rv:(\d+\.\d)/);

        /**
         * Определяет версию браузера Opera. Возвращает номер версии.
         * @name y5.Vars.opera_ver
         * @memberOf y5.Vars
         * @constant
         * @type Number
         */

        /**
         * @name y5.opera_ver
         * @link y5.Vars.opera_ver
         * @memberOf y5
         * @type Number
         */
        result.opera_ver    = getVer(result.is_opera, /opera[\/ ](\d+\.\d)/);

        /**
         * Определяет версию браузера Safari. Возвращает номер версии.
         * @name y5.Vars.safari_ver
         * @memberOf y5.Vars
         * @constant
         * @type Number
         */

        /**
         * @name y5.safari_ver
         * @link y5.Vars.safari_ver
         * @memberOf y5
         * @type Number
         */
        result.safari_ver   = getVer(result.is_safari, /safari\/(\d+)/);

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии 5.0.
         * @name y5.Vars.is_ie5
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie5
         * @link y5.Vars.is_ie5
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie5       = result.ie_ver == 5.0;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии 5.5.
         * @name y5.Vars.is_ie55
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie55
         * @link y5.Vars.is_ie55
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie55      = result.ie_ver == 5.5;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии 5.0 и последующих версий.
         * @name y5.Vars.is_ie5up
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie5up
         * @link y5.Vars.is_ie5up
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie5up     = result.ie_ver > 4.9;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии 5.5 и последующих версий.
         * @name y5.Vars.is_ie55up
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie55up
         * @link y5.Vars.is_ie55up
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie55up    = result.ie_ver > 5.4;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии 6.0 и последующих версий.
         * @name y5.Vars.is_ie6up
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie6up
         * @link y5.Vars.is_ie6up
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie6up     = result.ie_ver > 5.9;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии 7.0 и последующих версий.
         * @name y5.Vars.is_ie7up
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie7up
         * @link y5.Vars.is_ie7up
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie7up     = result.ie_ver > 6.9;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии ниже 6.0 (не включая 6.0).
         * @name y5.Vars.is_ie6down
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie6down
         * @link y5.Vars.is_ie6down
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie6down   = result.is_ie && result.ie_ver < 6.0;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии ниже 7.0 (не включая 7.0).
         * @name y5.Vars.is_ie7down
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie7down
         * @link y5.Vars.is_ie7down
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie7down   = result.is_ie && result.ie_ver < 7.0;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии ниже 8.0 (не включая 8.0).
         * @name y5.Vars.is_ie8down
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie7down
         * @link y5.Vars.is_ie8down
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie8down = result.is_ie && result.ie_ver < 8.0;

        /**
         * Определяет версию браузера. Возвращает true для браузеров MS Internet Explorer версии 8.
         * @name y5.Vars.is_ie8
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.is_ie55
         * @link y5.Vars.is_ie8
         * @memberOf y5
         * @type Boolean
         */
        result.is_ie8 = result.is_ie && result.ie_ver == 8;


        /**
         * Определяет доступность в браузере cookie. Устанавливается в true, если доступ к cookie возможен,  или в false, если cookie запрещены.
         * @name y5.Vars.cookieEnabled
         * @memberOf y5.Vars
         * @constant
         * @type Boolean
         */

        /**
         * @name y5.cookieEnabled
         * @link y5.Vars.cookieEnabled
         * @memberOf y5
         * @type Boolean
         */
        result.cookieEnabled = nav.cookieEnabled;

        return result;
    }
};

/// Копируем переменные браузера в Vars, затем в y5
(function () {
    var i;

    // Определение браузера
    var browser = y5.Browser.get(window.navigator);
    for (i in browser) {
        y5.Vars[i] = browser[i];
    }

    // Копируем переменные в y5
    for (i in y5.Vars) {
        y5[i] = y5.Vars[i];
    }

})();

/// Loader
/////////////////////////////////////////////////////////////////////

/**
 * @class Набор функций для загрузки дополнительных JavaScript-файлов и CSS-стилей на страницу.
 * @static
 * @name y5.Loader
 */
y5.Loader = (function() {

    function setElementAttributes(element, attributes) {
        for (var type in attributes) {
            var value = attributes[type];
            if (value) {
                element.setAttribute(type, value);
            }
        }
    }

    function insertElement(container, element) {
        container.insertBefore(element, container.firstChild);
    }

    var loadObject;
    if (y5.is_opera && y5.opera_ver < 8) {

        loadObject = function(name, attributes) {
            if (!document.body) {
                return null;
            }

            var span = document.createElement('span');
            span.style.display = 'none';
            span.innerHTML = '<' + name + '><\/' + name + '>';

            var element = span.getElementsByTagName(name).item(0);
            setElementAttributes(element, attributes);
            insertElement(document.body, span);

            return element;
        }

    } else {

        var head = document.getElementsByTagName('head')[0];

        loadObject = function(name, attributes) {
            var element = document.createElement(name);
            setElementAttributes(element, attributes);
            insertElement(head, element);

            return element;
        };

    }

    function createScript(attributes, callback) {
        var tries = 10, timer = null;

        function tryCreateScript() {
            if (!--tries) {
                window.clearTimeout(timer);
                return false;
            }

            var script = loadObject('script', attributes);

            if (script) {
                if (typeof callback == 'function') {
                    callback(script);
                }
                window.clearTimeout(timer);
                return true;
            }

            return false;
        }

        if (!tryCreateScript()) {
            timer = window.setInterval(tryCreateScript, 10);
        }
    }

    return {
        /**
         * Загружает файл, содержащий JavaScript
         * @name y5.Loader.loadScript
         * @memberOf y5.Loader
         * @function
         * @param {String} src Путь к скрипту
         * @param {String} [charset] Кодировка скрипта
         * @param {String} [callback] Функция, выполняемая по окончании загрузки
         * @param {String} [id] ID создаваемого элемента DOM
         */
        loadScript: function(src, charset, callback, id) {
            createScript({src: src, charset: charset, type: 'text/javascript', id: id}, callback);
        },

        /**
         * Загружает объект.
         * @name y5.Loader.loadObject
         * @memberOf y5.Loader
         * @param {String} name Имя элемента
         * @param {String} attributes Список атрибутов
         * @private
         *
         * @example
         * y5.Loader.loadObject('script', {src: 'script.js'});
         * // -> &lt;script src="script.js">&lt;/script>
         */
        loadObject: loadObject
    };

})();

/// Scripts
/////////////////////////////////////////////////////////////////////

/**
 * Создает элемент script.
 * @class
 * @static
 * @deprecated y5.Loader
 */
y5.Scripts = {
    /**
     * Создает элемент script.
     * @memberOf y5.Scripts
     * @deprecated y5.Loader.loadScript
     */
    createScript: function(src, charset, callback) {
        y5.Loader.loadScript(src, charset, callback);
    }
};

/// Console
/////////////////////////////////////////////////////////////////////

(function() {

var VOID = y5.VOID;

y5.Console = {
    log:      VOID,
    info:     VOID,
    warn:     VOID,
    error:    VOID,
    trace:    VOID,
    dir:      VOID,
    dirxml:   VOID,
    group:    VOID,
    groupEnd: VOID
};

})();

/// Exception
/////////////////////////////////////////////////////////////////////

/**
 * Создает экземпляр исключения
 * @class Класс Exception
 * @name y5.Exception
 * @param {String} message Сообщение
 * @param {String} method Название метода
 * @param {String} module Название модуля
 * @example
 * try {
 *     // JavaScript code
 * } catch (e) {
 *     throw new y5.Exception(e, 'myMethod', 'myModule');
 * }
 */
y5.Exception = function(message, method, module) {
    if (!y5.DEBUG) {
        return true;
    }

    var ret = 'y5.' + module + '.' + method + ': ' + message;
    var e = new Error(ret);

    if (e.stack) {
        e.message += "\nStack:\t" + e.stack.replace(/\n/ig, '\n\t');
    }

    return e;
};

y5.Exception.prototype = new Error();
