y5.require('Utils', function() {

var GC = y5.GC;

// вызов функции с контекстом
function execListenerWithContext(listener, context, param1, param2) {
    return (context ? listener.call(context, param1, param2) : listener(param1, param2));
}


// TimerObserver
/////////////////////////////////////////////////////////////////////

/**
 * Создает объект-таймер, вызывающий заданную функцию с заданной периодичностью.
 * @class Класс, реализующий механизм вызова заданных функций по таймеру.
 * @constructor
 * @name y5.TimerObserver
 * @param {Function} listener Callback-функция
 * @param {Number} period Периодичность запуска функции (в секундах)
 * @param {Boolean} add Начать слушать немедленно (по умолчанию - false)
 * @param {Object} context Контекст выполнения функции listener
 *
 * @example
 * function callback(observer) {
 *     // ...
 *     // останавливаем таймер после 10 запусков
 *     if (observer.tick == 10) {
 *         observer.remove();
 *     }
 * }
 * new y5.TimerObserver(callback, 5, true);
 */
y5.TimerObserver = function(listener, period, add, context) {
    this.period = period * 1000;
    this.added = false;
    this.tick = 0;
    this.timersCount = 10;
    this.timers = new Array(this.timersCount);
    this.context = context;
    this.listener = listener;

    if (add) {
        this.add();
    }

    GC.collect(this);
};

y5.TimerObserver.prototype = {
    /**
     * Количество вызовов функции.
     * @name y5.TimerObserver.tick
     * @memberOf y5.TimerObserver
     * @type Number
     */


    /**
     * Включает (активирует) таймер.
     * @name y5.TimerObserver.add
     * @memberOf y5.TimerObserver
     * @function
     */
    add: function() {
        if (this.added) {
            return;
        }

        this.tick = 0;
        this.added = true;
        this.setTimer();
    },

    /**
     * Выключает (деактивирует) таймер.
     * @name y5.TimerObserver.remove
     * @memberOf y5.TimerObserver
     * @function
     */
    remove: function() {
        if (!this.added) {
            return;
        }

        this.clearTimer();
        this.added = false;
    },

    /**
     * @private
     */
    execListener: function(startTime, i) {
        if (!this.added) {
            return;
        }

        window.clearTimeout(this.timers[i]);

        var endTime = new Date().getTime();
        var diff = startTime - endTime;
        var skip = diff > 0;

        this.tick++;
        execListenerWithContext(this.listener, this.context, this, this.tick, skip);

        if (this.tick % this.timersCount == 0) {
            this.setTimer();
        }
    },

    /**
     * @private
     */
    setTimer: function() {
        this.clearTimer();

        var period = 0;
        var time = new Date().getTime();
        for (var i = 0; i < this.timersCount; i++) {
            period += this.period;
            var startTime = time + period;

            this.timers[i] = y5.Utils.setTimeout(this.execListener, period, this, startTime, i);
        }
    },

    /**
     * @private
     */
    clearTimer: function() {
        for (var i = 0; i < this.timersCount; i++) {
            window.clearTimeout(this.timers[i]);
        }
    },

    /**
     * @private
     */
    cleanup: function() {
        this.remove();
        this.context = null;
        this.listener = null;
    }
};


/// InputObserver
/////////////////////////////////////////////////////////////////////

/**
 * Создает объект-слушатель.
 * @class Класс для создания объектов-слушателей, отслеживающих изменения элементов ввода. Предназначен для создания интерактивных (реагирующих на заполнение пользователем полей ввода) HTML-форм.
 * @constructor
 * @name y5.InputObserver
 * @param {Function} listener Callback-функция
 * @param {Element} element Элемент
 * @param {Boolean} [add] Начать слушать немедленно (по умолчанию - false)
 * @param {Object} [context] Контекст выполнения функции listener
 *
 * @example
 * new y5.InputObserver(callback, element, true);
 */
y5.InputObserver = function(listener, element, add, context) {
    this.element = element;
    this.listener = listener;
    this.context = context;
    this.added = false;

    // init
    this.setInitState();
    this.timer = new y5.TimerObserver(this.execListener, .1, this.added, this);

    if (add) {
        this.add();
    }

    GC.collect(this);
};

y5.InputObserver.prototype = {
    /**
     * Включает (активирует) объект-слушатель.
     * @name y5.InputObserver.add
     * @memberOf y5.InputObserver
     * @function
     */
    add: function() {
        this.timer.add();
        this.added = true;
    },

    /**
     * Выключает (деактивирует) объект-слушатель.
     * @name y5.InputObserver.remove
     * @memberOf y5.InputObserver
     * @function
     */
    remove: function() {
        this.timer.remove();
        this.added = false;
    },

    /**
     * @private
     */
    execListener: function(timer, skip) {
        if (/* TEST: !skip && */this.isChanged()) {
            execListenerWithContext(this.listener, this.context, this.element);
            this.state = this.getState();
        }
    },

    /**
     * Запрашивает текущее состояние элемента ввода. Возвращает объект состояния.
     * @name y5.InputObserver.getState
     * @memberOf y5.InputObserver
     * @function
     * @returns {Object} Объект состояния
     */
    getState: function() {
        var element = this.element;

        return {
            value: element.value,
            checked: element.checked,
            selected: element.selected,
            selectedIndex: element.selectedIndex
        };
    },

    /**
     * Устанавливает текущее состояние элемента в качестве начального.
     * @name y5.InputObserver.setInitState
     * @memberOf y5.InputObserver
     * @function
     */
    setInitState: function() {
        this.state = this.getState();
        this.initState = this.getState();
    },

    /**
     * Возвращает изменение состояния ввода в сравнении с предыдущим.
     * @private
     * @returns {Boolean} Изменение текущего состояния элемента в сравнении с this.state
     * @note Изменение this.state происходит при каждом выполнении функции callback.
     */
    isChanged: function() {
        return this.stateChanged(this.state);
    },

    /**
     * Проверяет, изменилось ли состояние элемента ввода по сравнению с начальным. Если состояние изменилось, то возвращается true, иначе - false.
     * @name y5.InputObserver.isInitChanged
     * @memberOf y5.InputObserver
     * @function
     * @returns {Boolean} Изменение текущего состояния элемента в сравнении с this.initState
     */
    isInitChanged: function() {
        return this.stateChanged(this.initState);
    },

    /**
     * Сравнивает состояния элемента ввода.
     * @private
     * @param {Object} state Сравниваемое состояние
     * @returns {Boolean} Изменение текущего состояния элемента в сравнении с state
     */
    stateChanged: function(state) {
        var element = this.element;

        return (
            state.checked != element.checked ||
            state.selected != element.selected ||
            state.selectedIndex != element.selectedIndex ||
            state.value != element.value
        );
    },

    /**
     * @private
     */
    cleanup: function() {
        this.remove();
        this.context = null;
        this.listener = null;
        this.element = null;
    }
};


/// Events Factory
/////////////////////////////////////////////////////////////////////

/**
 * Создает объект-слушатель, реагирующий на изменение состояния элемента ввода.
 * @name y5.Events.observeInput
 * @memberOf y5.Events
 * @function
 * @param {Function} listener Callback-функция
 * @param {Element} element Элемент
 * @param {Boolean} [add] Начать слушать немедленно (по умолчанию - false)
 * @param {Object} [context] Контекст выполнения функции listener
 * @returns {y5.InputObserver}
 *
 * @example
 * y5.Events.observeInput("value", callback, element, true);
 */
y5.Events.observeInput = function(listener, element, add, context) {
    return new y5.InputObserver(listener, element, add, context);
};

/**
 * Позволяет организовать вызов функции по таймеру.
 * @name y5.Events.observeTimer
 * @memberOf y5.Events
 * @function
 * @param {Function} listener Callback-функция
 * @param {Number} period Периодичность запуска функции (в секундах)
 * @param {Boolean} [add] Начать слушать немедленно (по умолчанию - false)
 * @param {Object} [context] Контекст выполнения функции listener
 * @returns {y5.TimerObserver}
 *
 * @example
 * function callback(observer) {
 *     // ...
 *     // останавливаем таймер после 10 запусков
 *     if (observer.tick == 10) {
 *         observer.remove();
 *     }
 * }
 * y5.Events.observeTimer(callback, 5, true);
 */
y5.Events.observeTimer = function(listener, period, add, context) {
    return new y5.TimerObserver(listener, period, add, context);
};

y5.loaded('EventsExt');

});