y5.require('Strings', function() {

var stringTestRegexp = /[\x00-\x1f\"\\]/,
    stringReplaceRegexp = /[\x00-\x1f\"\\]/g,
    compositeKeyTestRegexp = /^(.+)\[(.*)\]$/,
    specialChars = {
        '\b': '\\b',
        '\t': '\\t',
        '\n': '\\n',
        '\f': '\\f',
        '\r': '\\r',
        '"' : '\\"',
        '\\': '\\\\'
    },
    Types = y5.Types,
    Strings = y5.Strings;

/**
 * Функция преобразования специальных и не-ASCII символов
 * в escape-последовательности для представления в JSON.
 * @param {String} char символ для преобразования
 * @returns {String} escape-последовательность
 * @private
 */
function replaceChars(ch) {
    var chr = specialChars[ch];
    if (chr) return chr;

    chr = ch.charCodeAt();
    return '\\u00' + Math.floor(chr / 16).toString(16) + (chr % 16).toString(16);
}

/**
 * Функция форматирования числа в двузначную строку.
 * @param {Number} number Число
 * @returns {String} Строка
 * @private
 *
 * @example
 * 1 -> '01'
 * 10 -> '10'
 */
function formatDateItem(num) {
    if (num < 0) {
        num = 0;
    }
    return num < 10 ? '0' + num : num;
}

/**
 * @class Набор функций для работы с данными в формате JSON.
 * @static
 * @name y5.JSON
 * @memberOf y5
 */
y5.JSON = {
    /**
     * Преобразовывает (сериализует) JavaScript-объект в JSON-строку. Возвращает JSON-строку.
     * @name y5.JSON.encode
     * @memberOf y5.JSON
     * @function
     * @param {Object} object Объект для сериализации
     * @returns {String} Сериализованный объект
     *
     * @example
     * y5.JSON.encode({a: 1, b: true, c: 'foo'});
     * // -> '{"a":1,"b":true,"c":"foo"}'
     * y5.JSON.encode([1, 2, 3]);
     * // -> '[1,2,3]'
     */
    encode: function(obj) {
        switch (Types.type(obj)) {
            case Types.NUMBER:
                return this.fromNumber(obj);

            case Types.BOOLEAN:
                return this.fromBoolean(obj);

            case Types.STRING:
                return this.fromString(obj);

            case Types.DATE:
                return this.fromDate(obj);

            case Types.ARRAY:
                return this.fromArray(obj);

            case Types.OBJECT:
                return this.fromObject(obj);

            default:
                return 'null';
        }
    },

    /**
     * Преобразовывает JSON-строку в JavaScript-объект. Возвращает объект, полученный из JSON.
     * @name y5.JSON.decode
     * @memberOf y5.JSON
     * @function
     * @param {String} string Строка JSON
     * @returns {Object} Объект
     * 
     * @example
     * y5.JSON.decode('{"a":1,"b":true,"c":"foo"}');
     * // -> {a: 1, b: true, c: "foo"} 
     */
    decode: function(s) {
        return eval('(' + s + ')');
    },
    
    /**
     * Возвращает JSON-строку, полученную из URL
     * @param {String} query Строка запроса
     * @returns {String} Строка JSON
     * 
     * @example
     * y5.JSON.fromQuery('?test=1&amp;&amp;test=2');
     * // -> {"test":"2"}
     * 
     * y5.JSON.fromQuery('test=foo&amp;foo[14]=3&amp;foo[]=4&amp;foo[]=5&amp;foo[test]=2test1=text');
     * // -> {"test":"foo","foo":{14:"3",15:"4",16:"5","test":"2test1=text"}}
     */
    fromQuery: function (query) {
        query = query.toString() || '';
        // выбирезаем запрос после ?
        var queryIndex = query.indexOf('?');
        if (queryIndex != -1) {
            query = query.substr(queryIndex+1);
        }
        
        if (query) {
            query = query.split('&');
            var i=0,
                queryLength = query.length,
                result = {},
                compositeKey,
                compositeIndex,
                compositeObject,
                value,
                key,
                vars;

            for (;i<queryLength;i++) {
                if (!query[i]) continue;
                queryIndex = query[i].indexOf('=');
                if (queryIndex == -1) {
                    key = query[i];
                    value = '';    
                } else {
                    key = query[i].substring(0, queryIndex);
                    value = query[i].substr(queryIndex+1) || '';
                }
                
                // проверяем элементы типа foo[]
                compositeKey = key.match(compositeKeyTestRegexp);
                
                if (compositeKey && compositeKey[1]) {
                    // если такого элемента нет, то надо его создать
                    if (!Types.object(result[compositeKey[1]])) {
                        result[compositeKey[1]] = {};
                        //фейковый параметр для того, чтобы работало foo[13]=13,foo[test]=test,foo[]=14 
                        result[compositeKey[1]].__length = 0;
                    }
                    compositeObject = result[compositeKey[1]];
                    
                    // если указан индекс объекта foo[index] 
                    if (compositeKey[2]) {
                        compositeIndex = parseInt(compositeKey[2], 10);
                        // если индекс числовой
                        if (!isNaN(compositeIndex)) {
                            compositeObject[compositeIndex] = value;
                            if (compositeIndex > compositeObject.__length) {
                                compositeObject.__length = compositeIndex+1;
                            }
                                
                        } else if (Types.string(compositeKey[2])) {
                            compositeObject[compositeKey[2]] = value;
                        }
                        
                    // если индекс пустой, т.е. foo[], то как бы делаем push
                    } else {
                        compositeObject[compositeObject.__length++] = value;
                    }        
                            
                } else {
                    result[key] = value;    
                }
            }
            
            // удаляем __length
            for (i in result) {
                if (result[i] && result[i].__length) {
                    delete result[i].__length;
                }
            }
            
            return this.encode(result);
        }
        return 'null';
    },
    
    /**
     * Преобразует JavaScript-объект в строку запроса
     * @param {Object} object Объект
     * @returns {String} Строка запроса
     * 
     * @example
     * y5.JSON.toQuery({"test":"foo","foo":{14:"3",15:"4",16:"5","test":"2test1=text"}});
     * // -> test=foo&amp;foo[14]=3&amp;foo[15]=4&amp;foo[16]=5&amp;foo[test]=2test1=text
     * 
     * y5.JSON.toQuery({array:[1,2,3]});
     * // -> array[]=1&amp;array[]=2&amp;array[]=3
     * 
     * y5.JSON.toQuery({dt:new Date(2008,06,13,14,15,03)});
     * // -> dt=2008-07-13T10:15:03Z
     */
    toQuery: function (object) {
        var query = '',
            i,
            delimeter = '',
            index,length,obj;

        object = Types.object(object)?object:this.decode(object);
            
        for (i in object) {
            obj = object[i];
            switch (Types.type(obj)) {
                case Types.BOOLEAN:
                    query += delimeter + i+'='+this.fromBoolean(obj);
                    delimeter = '&';
                    break;
    
                case Types.DATE:
                    query += delimeter + i+'='+this.fromDate(obj).replace('"', '').replace('"', '');
                    delimeter = '&';
                    break;
    
                case Types.ARRAY:
                    for (index=0,length=obj.length; index<length; index++) {
                        query += delimeter + i + '[]=' + obj[index];
                        delimeter = '&';
                    }
                    break;
    
                case Types.OBJECT:
                    for (index in obj) {
                       query += delimeter + i + '[' + index + ']=' + obj[index];
                        delimeter = '&';
                    }
                    break;
    
                default:
                    query += delimeter + i+'='+obj;
                    delimeter = '&';
                    break;
            }
        }
        return query;
    },
    
    /**
     * Преобразовывает JSON-строку в формат, пригодный для вставки в HTML-код страницы. Возвращает строку, в которой все служебные HTML-символы заменены на символьные примитивы. Полезно для визуального контроля содержимого JSON при отладке приложения.
     * @name y5.JSON.toHTML
     * @memberOf y5.JSON
     * @function
     * @param {Object} object Объект для сериализации
     * @returns {String} Сериализованный объект
     *
     * @example
     * y5.JSON.toHTML({a: 'foo&amp;bar'});
     * // -> '{&amp;quot;a&amp;quot;:&amp;quot;foo&amp;amp;bar&amp;quot;}'
     */
    toHTML: function(obj) {
        return Strings.escapeHTML(this.encode(obj));
    },

    /**
     * Возвращает строковое представление числа для JSON.
     * @param {Number} object Объект для сериализации
     * @returns {String} Сериализованный объект
     * @private
     */
    fromNumber: function(obj) {
        return isFinite(obj) ? String(obj) : 'null';
    },

    /**
     * Возвращает строковое представление булевого объекта для JSON.
     * @param {Boolean} object Объект для сериализации
     * @returns {String} Сериализованный объект
     * @private
     */
    fromBoolean: function(obj) {
        return String(obj);
    },

    /**
     * Возвращает строку для JSON.
     * Специальные и не-ASCII символы заменяются escape-последовательностями.
     * @param {String} object Объект для сериализации
     * @returns {String} Сериализованный объект
     * @private
     */
    fromString: function(obj) {
        if (stringTestRegexp.test(obj)) {
            try {
                obj = obj.replace(stringReplaceRegexp, replaceChars);
            } catch (e) {}
        }

        return '"' + obj + '"';
    },

    /**
     * Возвращает строковое представление даты для JSON.
     * @param {Date} object Объект для сериализации
     * @returns {String} Сериализованный объект
     * @private
     */
    fromDate: function(obj) {
        return '"' + obj.getUTCFullYear()         + '-' +
            formatDateItem(obj.getUTCMonth() + 1) + '-' +
            formatDateItem(obj.getUTCDate())      + 'T' +
            formatDateItem(obj.getUTCHours())     + ':' +
            formatDateItem(obj.getUTCMinutes())   + ':' +
            formatDateItem(obj.getUTCSeconds())   + 'Z"';
    },

    /**
     * Возвращает строковое представление массива для JSON.
     * @param {Array} object Объект для сериализации
     * @returns {String} Сериализованный объект
     * @private
     */
    fromArray: function(obj) {
        return '[' + obj.map(function(item) { return this.encode(item) }, this).join(',') + ']';
    },

    /**
     * Возвращает JSON-строку.
     * @param {Object} object Объект для сериализации
     * @returns {String} Сериализованный объект
     * @private
     */
    fromObject: function(obj) {
        var list = [];

        for (var i in obj) {
            if (i == Number(i)) {
                i = Number(i);
            }
            list.push(this.encode(i) + ':' + this.encode(obj[i]));
        }

        return '{' + list.join(',') + '}';
    }
};

var JSON = y5.JSON;

/// aliases

/**
 * @name y5.JSON.toObject
 * @function
 * @see y5.JSON.decode
 */
JSON.toObject = JSON.decode;

/**
 * @name y5.JSON.toString
 * @function
 * @see y5.JSON.encode
 */
JSON.toString = JSON.encode;

y5.loaded('JSON');

});