y5.require('Strings', 'Events', 'URL', 'Utils', function() {

var URL = y5.URL,
    Utils = y5.Utils,
    Types = y5.Types,
    VOID = y5.VOID,
    moduleName = 'Request',
    states = ['uninitialized', 'loading', 'loaded', 'interactive', 'complete'];

function onExceptionDefault(e) {
    y5.Console.error(moduleName, [e], [moduleName]);
}

function trimText() {
    return y5.Strings.trim(this.responseText);
}

function responseText() {
    return this.responseText;
}

function jsonDecode() {
    if (y5.JSON) {
        return y5.JSON.decode(this.responseText);
    }
    throw new y5.Exception('y5:JSON module required', jsonDecode, 'Request');
}

/**
 * Создает базовый экземпляр класса для реализации на его основе Ajax-запросов.
 * @name y5.Request
 * @class Набор базовых методов для реализации на их основе Ajax-запросов с использованием разных методов передачи данных (транспортов): XML, Script, Iframe.
 * @constructor
 * @param {String | y5.URL} url URL запроса
 * @param {Object} params Параметры запроса
 *
 * @example
 * // Пример запроса (на примере y5.Request.XML).
 * var req = new y5.Request.XML("test.xml");
 * req.onload = function(req) { alert(req.responseText) };
 * req.onexception = function(e) { alert(e.message) };
 * req.send();
 *
 * // Запрос, обработчики которого находятся в объекте
 * var obj = {
 *     errorMessage: 'Ошибка загрузки',
 *     onload: function(req) { alert(req.responseText) },
 *     onerror: function() { alert(this.errorMessage) }
 * };
 *
 * var req = new y5.Request.XML("test.xml", { callbackObject: obj });
 * req.send();
 *
 *
 * Типы событий на выполнение запроса:
 *
 * loading: возникает при отправке запроса
 * load: возникает при успешном завершении запроса и получении запрашиваемых данных
 * error: ошибка при выполнении запроса
 * abort: возникает при прерывании запроса с помощью функции abort()
 * exception: ошибка связанная с работой приложения
 * complete: возникает всегда по завершении запроса, если не возникло исключения
 * {код ответа сервера}: например '200'. Если сервер вернул определенный статус и существует
 * функция такого вида, то она будет вызвана
 *
 *
 * Функции, реагирующие на события можно привязывать разными способами:
 *
 * 1) Непосредственное назначение функции экземпляру объекта (имя функции -- on{тип события}).
 * 2) Передача в виде параметров при создании экземпляра класса ({onload: function() {}})
 * 3) Навешивание функций на события. При возникновении события посылается сообщение
 *    типа request:{тип события} для ID указанного в параметрах при создании экземпляра.
 *
 *
 * Объект, возвращаемый при получении запроса имеет следующие свойства:
 *
 * responseText: тело ответа
 * text(): функция возвращает responseText с удаленными начальными и концевыми пробельными символами
 * getResponseHeader(header): возвращает заголовок ответа
 * status: код ответа (если статус в диапазоне от 200 до 299, то возникает событие load, иначе error)
 */
function Request(url, params) {
    this.req = null;
    this.url = url instanceof URL ? url : new URL(url);

    var context;
    if (params) {
        // получаем ссылку на контекст выполнения обработчиков для того, чтобы не копировать его через objectCopy
        context = params.callbackContext;
        params.callbackContext = null;
    }

    // параметры
    this.params = Utils.objectCopy(
        Utils.objectCopy({}, this.defaultParams),
        params
    );

    if (Types.object(context)) {
        this.params.callbackContext = context;
    }

    this.params.method = this.params.method.toLowerCase();
}

Request.prototype = {
    /**
     * Параметры по умолчанию.
     * @memberOf y5.Request
     * @name y5.Request.params
     * @private
     * @type Object
     *
     * @example
     * // параметры по умолчанию
     * {
     *     // идентификатор запроса
     *     id: null,
     *
     *     // метод запроса
     *     method: 'get',
     *
     *     // функция выполняемая при исключении
     *     onexception: onExceptionDefault,
     *
     *     // контекст выполнения функций callback
     *     callbackContext: null,
     *
     *     // объект содержащий функции callback
     *     callbackObject: null
     * }
     */
    defaultParams: {
        // идентификатор запроса
        id: null,

        // метод запроса
        method: 'get',

        // функция выполняемая при исключении
        onexception: onExceptionDefault,

        // контекст выполнения функций callback
        callbackContext: null,

        // объект содержащий функции callback
        callbackObject: null
    },

    /**
     * Прерывает выполнение запроса. Возникает событие abort.
     * @memberOf y5.Request
     * @name y5.Request.abort
     * @function
     */
    abort: VOID,

    /**
     * Отправляет запрос. В качестве аргумента может быть передана строка в виде параметров GET или объект JavaScript.
     * @memberOf y5.Request
     * @name y5.Request.send
     * @function
     * @param {String | Object} [query] Данные запроса в виде запроса GET или объекта
     *
     * @example
     * req.send('foo=1&amp;bar=2');
     * req.send({foo: 1, bar: 2});
     */
    send: function(query) {
        try {
            this.init();
            y5.Console.log(this.toString() + ' ' + this.params.method.toUpperCase(), [this, query], [moduleName, this.toString()]);

            if (query && y5.Types.object(query) && !query.submit) {
                query = y5.Url('').replaceParams(query).query();
            }

            this._send(query);
        } catch (e) {
            this.dispatch('exception', e);
        }
    },

    /**
     * @private
     */
    init: function() {
        if (this.isInit) return;

        if (!this.id) {
            this.id = Utils.generateUniqueId();
        }
        this._init();

        this.isInit = true;
    },

    /**
     * @private
     */
    _init: VOID,

    /**
     * @private
     */
    end: function() {
        this._end();

        // clean up to avoid memleak
        this.req = null;
        this.isInit = false;

        this.dispatch('end');
    },

    /**
     * @private
     */
    _end: VOID,

    /**
     * @private
     */
    dispatch: function(status, args) {
        var params = this.params,
            name = 'on' + status,
            callbackContext = params.callbackContext || this,
            callbackObject = params.callbackObject;

        args = args || this.req;

        // выполняем функцию
        if (Types.func(this[name])) {
            this[name].call(callbackContext, args);
        }

        // выполняем функцию, если задана в параметрах
        if (Types.func(params[name])) {
            params[name].call(callbackContext, args);
        }

        // выполняем функцию объекта содержащего функции callback
        if (callbackObject && Types.func(callbackObject[name])) {
            callbackObject[name](args);
        }

        // посылаем сообщение
        y5.Notify('request:' + status, params.id || this, args);
        
        // глобальное отслеживание запросов
        y5.Notify('request:' + status, y5, [this, args]);
    },

    /**
     * @private
     */
    onStateChange: function() {
        try {
            y5;
        } catch (e) {
            // обрыв запроса при закрытии страницы
            return;
        }

        var stateNum = this.req.readyState,
            complete = stateNum == 4;

        if (complete) {
            var status = 0;
            try {
                status = this.req.status;
            } catch (e) {}

            if (y5.is_ie && status == 1223) {
                // http://dev.jquery.com/ticket/1450
                status = 204
            }

            if (!Types.func(this['on' + status])) {
                status = (status >= 200 && status < 300) ? 'load' : 'error';
            }

            var isText = Types.string(this.req.responseText);
            this.req.text = isText ? trimText : responseText;
            this.req.json = isText ? jsonDecode : y5.NULL;

            // выполняем функцию по окончании загрузки
            this.dispatch(status);
        }

        // выполняем функцию на изменение состояния загрузки
        this.dispatch(states[stateNum]);

        if (complete) {
            this.end();
        }
    },

    /**
     * Возвращает название класса.
     * @memberOf y5.Request
     * @name y5.Request.toString
     * @function
     * @returns {String} Название класса
     */
    toString: function() {
        return moduleName;
    }
};

y5.Get = {};
y5.Post = {};
y5.Load = {};

/**
 * Функция для расширения возможностей конкретных Ajax-классов.
 * Добавляет функции: y5.Get.*, y5.Post.*, y5.Load.*.
 * @private
 */
Request.ext = function(type, obj, notExt) {
    if (Types.undef(notExt)) {
        Utils.objectExtends(obj, Request, moduleName);
    }

    var module = moduleName + '.' + type;

    y5.Get[type] = function(url, params, query) {
        var xhr = new obj(url, Utils.objectCopy(params || {}, {method: 'get'}));
        xhr.send(query);
        return xhr;
    };

    y5.Post[type] = function(url, params, query) {
        var xhr = new obj(url, Utils.objectCopy(params || {}, {method: 'post'}));
        xhr.send(query);
        return xhr;
    };

    y5.Load[type] = function(container, url, params, query) {
        var xhr = new obj(url, params);
        xhr.onload = function(req) {
            if (Types.string(container)) {
                container = document.getElementById(container);
            }
            container.innerHTML = req.responseText;
        };
        xhr.send(query);
        return xhr;
    };

    obj.prototype.toString = function() {
        return module;
    };

    Request[type] = obj;

    y5.loaded(module);
};

y5.Request = Request;
y5.loaded(moduleName);

});