y5.require(['Dom', 'Events'], function() {

/**
 * Класс, реализующий работу со слушателями "горячих клавиш".
 * @class Содержит функции для работы со слушателями "горячих клавиш".
 * @public
 * @name y5.ShortCutListener
 * @memberOf y5
 * @param {Number} masks Маски комбинаций клавиш
 * @param {Function} callback Функция обработки нажатия клавиш
 * @param {Node} node Целевой элемент
 * @param {Object} options Параметры
 */
function ShortCutListener(masks, callback, node, options) {
    this.masks = masks;
    this.node = node;

    if (options.context) {
        this.callback = function(e, opts) {
            callback.apply(options.context, [e, opts]);
        }
    } else {
        this.callback = callback;
    }
    this.options = options;

    this.add();

    y5.GC.collect(this);
}

ShortCutListener.prototype = {
    /// Public

    /**
     * Добавляет комбинацию клавиш в список активных.
     * @name y5.ShortCutListener.add
     * @memberOf y5.ShortCutListener
     * @public
     * @function
     */
    add: function() {
        this.enable(true);
    },

    /**
     * Удаляет комбинацию клавиш из списка активных.
     * @name y5.ShortCutListener.remove
     * @memberOf y5.ShortCutListener
     * @public
     * @function
     */
    remove: function() {
        this.enable(false);
    },

    /**
     * Проверяет активность комбинации.
     * @name y5.ShortCutListener.isEnable
     * @memberOf y5.ShortCutListener
     * @public
     * @function
     */
    isEnable: function() {
        return this.enabled;
    },

    /**
     * Включает/выключает состояние активности.
     * @name y5.ShortCutListener.enable
     * @memberOf y5.ShortCutListener
     * @public
     * @function
     * @param {Boolean} enabled
     */
    enable: function(enabled) {
        this.enabled = enabled;
    },

    /// Private

    /**
     * Проверяет, нажата ли комбинация и если да, то вызывает callback.
     * @param {Number} mask Маска комбинации клавиш
     * @param {Boolean} isInput Является ли элемент элементом ввода?
     * @param {Node} node Элемент, на котором произошло событие
     * @param {Event} event Событие ввода
     * @private
     */
    check: function(mask, isInput, node, e) {
        if (!this.enabled) {
            return false;
        }

        // выходим, если сокращение запрещено на элементах ввода
        if (this.options.checkTarget && isInput) {
            return false;
        }

        // попали в маску?
        if (!this.checkMask(mask)) {
            return false;
        }

        // проверяем, если элемент на котором произошло событие
        // является дочерним по отношению к целевому элементу
        if (y5.Dom.isChild(node, this.node)) {
            if (this.options.preventDefault) {
                e.preventDefault();
            }

            this.callback(e, this.options);
            return true;
        }

        return false;
    },

    /**
     * Проверяет маску по списку масок объекта.
     * @param {Number} mask Маска комбинации клавиш
     * @param {Boolean} Результат проверки
     * @private
     */
    checkMask: function(mask) {
        for (var i = 0, l = this.masks.length; i < l; i++) {
            var m = this.masks[i];
            if (m == mask || m == 0) {
                return true;
            }
        }
        return false;
    },

    /**
     * Очищает содержимое этого объекта.
     * @name y5.ShortCutListener.cleanup
     * @memberOf y5.ShortCutListener
     * @public
     * @function
     */
    cleanup: function() {
        this.remove();
        this.node = null;
        this.options = null;
        this.callback = null;
    }
};

/// ShortCut

var maskUnicode = 1 << 16,
    maskCode = maskUnicode - 1,
    maskCtrl = maskUnicode,
    maskAlt = maskCtrl << 1,
    maskShift = maskAlt << 1,
    typeDown = 1,
    typePress = 2,
    default_options = {
        checkTarget: true,
        preventDefault: true,
        context: null,
        once: false
    },
    listDown = [],
    listPress = [];

var KEYS = {
    BS: 8,
    BACKSPACE: 8,
    TAB: 9,
    ENTER: 13,
    SHIFT: 16,
    CTRL: 17,
    ALT: 18,
    PAUSE: 19,
    CAPS_LOCK: 20,
    ESC: 27,
    SPACE: 32,
    PAGE_UP: 33,
    PGUP: 33,
    PAGE_DOWN: 34,
    PGDN: 34,
    END: 35,
    HOME: 36,
    LEFT_ARROW: 37,
    LEFT: 37,
    UP_ARROW: 38,
    UP: 38,
    RIGHT_ARROW: 39,
    RIGHT: 39,
    DOWN_ARROW: 40,
    DOWN: 40,
    INSERT: 45,
    INS: 45,
    DELETE: 46,
    DEL: 46,
    LEFT_WINDOW: 91,
    RIGHT_WINDOW: 92,
    SELECT: 93,
    PLUS: y5.is_ie || y5.is_safari ? 187 : 61,
    PLUS_NUM: 107,
    MINUS: y5.is_ie || y5.is_safari ? 189 : 109,
    MINUS_NUM: 109,
    NUM_1: 49,
    F1: 112,
    F2: 113,
    F3: 114,
    F4: 115,
    F5: 116,
    F6: 117,
    F7: 118,
    F8: 119,
    F9: 120,
    F10: 121,
    F11: 122,
    F12: 123,
    NUM_LOCK: 144,
    SCROLL_LOCK: 145,
    SLASH: 191,
    ASTERISK: 106
};

function convertKey(key) {
    if (y5.Types.object(key)) {
        return key;
    }

    var tokens = key.replace(/\s+/g, '').split(/\+/g);
    key = {};

    tokens.forEach(function(tok) {
        tok = tok.toUpperCase();
        switch (tok) {
            case 'ALT':
                key.alt = true;
                break;

            case 'SHFT': case 'SHIFT':
                key.shift = true;
                break;

            case 'CTL': case 'CTRL':
                key.ctrl = true;
                break;

            default:
                var num = KEYS[tok];
                if (num) {
                    key.key = num;
                } else {
                    key.ch = tok;
                }
        }
    });

    return key;
}

function convertKeys(keys) {
    if (!y5.Types.array(keys)) {
        keys = [keys];
    }

    return keys.map(convertKey);
}

function setDefaultOption(options) {
    for (var option in default_options) {
        if (typeof options[option] === y5.UNDEF) {
            options[option] = default_options[option];
        }
    }

    return options;
}

function removeFromList(element, list) {
    var index = list.lastIndexOf(element);
    if (index !== -1) {
        y5.GC.remove(list[index]);
        list = list.splice(index, 1);
        return true;
    }
    return false;
}

function getMask(set, type) {
    var code = 0, mask;

    if (set.key) {
        code = set.key;
    } else if (set.ch) {
        switch (type) {
            case typeDown:
                code = set.ch.toUpperCase();
                break;

            case typePress:
                code = set.ch.toLowerCase();
                break;
        }
        code = code.charCodeAt(0);
    }

    mask = maskCode & code;

    if (set.ctrl) {
        mask ^= maskCtrl;
    }

    if (set.alt) {
        mask ^= maskAlt;
    }

    if (set.shift) {
        mask ^= maskShift;
    }

    return mask;
}

function getMasks(keys, type) {
    var len = keys.length,
        masks = new Array(len);

    for (var i = 0; i < len; i++) {
        masks[i] = getMask(keys[i], type);
    }

    return masks;
}

function getSet(key, e) {
    return {key: key, ctrl: e.ctrlKey, alt: e.altKey, shift: e.shiftKey}
}

function getMaskEvent(e, type) {
    switch (type) {
        case typeDown:
            return getMask(getSet(e.keyCode, e));

        case typePress:
            var key = e.charCode ? e.charCode : e.keyCode;
            return getMask(getSet(key, e));
    }
    return 0;
}

function isInputTarget(elem) {
    if (!elem.tagName) {
        return false;
    }

    switch (elem.tagName.toLowerCase()) {
        case 'input': {
            switch (elem.type) {
                case 'text': case 'password': case 'file': case 'search':
                    return true;
            }
            break;
        }
        case 'textarea': {
            return true;
        }
    }

    return false;
}

function keyListener(e, type, list) {
    var i, l,
        node = e.target,
        mask = getMaskEvent(e, type),
        isInput = isInputTarget(node),
        tmpList = [],
        res = false;

    for (i = 0, l = list.length; i < l; i++) {
        if (list[i].isEnable()) {
            tmpList.push(list[i]);
        }
    }

    for (i = 0, l = tmpList.length; i < l; i++) {
        var shortcut = tmpList[i];
        var check = shortcut.check(mask, isInput, node, e);

        if (check && shortcut.options.once) {
            removeFromList(shortcut, list);
        }

        res = check || res;
    }

    return res;
}

function keyDownListener(e) {
    return keyListener(e, typeDown, listDown);
}

function keyPressListener(e) {
    // блокируем сочетания, на которые могут быть повешены браузерные сочетания
    if (!(e.ctrlKey || e.altKey)) {
        e.stopPropagation();
    }

    return keyListener(e, typePress, listPress);
}

function action(type, list, keys, listener, node, options) {
    switch (typeof options) {
        case 'object':
            break;

        case 'boolean':
            options = {checkTarget: options};
            break;

        case 'undefined':
            options = {};
            break;
    }

    options = setDefaultOption(options);

    return (list[list.length] = new ShortCutListener(getMasks(convertKeys(keys), type), listener, node || document, options));
}

/**
 * @class Набор функций для обработки клавиатурных сокращений. Позволяет реализовать в приложении поддержку «горячих клавиш».
 * @static
 * @name y5.ShortCut
 *
 * @example
 * Предопределенные константы клавиш:
 *
 *   F1  BACKSPACE END          PLUS_NUM
 *   F2  TAB       HOME         MINUS
 *   F3  ENTER     LEFT_ARROW   MINUS_NUM
 *   F4  SHIFT     UP_ARROW     NUM_1
 *   F5  CTRL      RIGHT_ARROW  NUM_LOCK
 *   F6  ALT       DOWN_ARROW   SCROLL_LOCK
 *   F7  PAUSE     INSERT       SLASH
 *   F8  CAPS_LOCK DELETE       ASTERISK
 *   F9  ESC       PLUS         BS
 *   F10 SPACE     LEFT_WINDOW
 *   F11 PAGE_UP   RIGHT_WINDOW
 *   F12 PAGE_DOWN SELECT
 *
 * Вызов функций:
 * y5.ShortCut.down([Key, Key], Callback, Node, Options);
 *
 * Примеры сочетаний:
 * 'h'
 * 'ctrl+h'
 * ['ctrl+h','alt+b']
 * {ctrl: true, ch: 'h'}
 * {ctrl: true, key: y5.ShortCut.ESC}
 * [{ctrl: true, ch: 'h'}, {alt: true, ch: 'b'}]
 * ['ctl+h', {alt: true, ch: 'b'}]
 * 'alt+f1'
 * 'ctrl + alt + shift + h'
 *
 * Пример вызова:
 * y5.ShortCut.down([{ch: 'a', ctrl: true}], function() { alert('Ctrl+A'); }, document);
 * y5.ShortCut.down([{key: y5.ShortCut.ESC}], function() { alert('Esc'); }, document);
 *
 * Объект Key:
 * {
 *    key: code,              // числовой код клавиши
 *    ch: char,               // символ клавиши, например 'A'
 *    ctrl: (true | false),   // зажата клавиша Ctrl
 *    alt: (true | false),    // зажата клавиша Alt
 *    shift: (true | false),  // зажата клавиша Shift
 * }
 *
 * Если указано свойство key, то свойство ch игнорируется.
 * Свойства ctrl, alt, shift являются необязательными.
 *
 * Объект Options:
 * {
 *    checkTarget: (true | false),     // проверять целевой элемент (если true (по умолчанию),
 *                                     // то не вызывать Callback на полях ввода)
 *    preventDefault: (true | false),  // запрещать выполнять действие по умолчанию, которое
 *                                     // установлено для документа на нажатие клавиши (true по умолчанию)
 *    context: (null | context),       // контекст в котором будет выполнена функция Callback
 *                                     // (контекста по умолчанию нет)
 *    once: (false | true)             // обработать сочетание один раз, затем удалить (false по умолчанию)
 * }
 */
y5.ShortCut = {
    /**
     * Устанавливает обработчик клавиатурного сокращения на нажатие клавиши (keydown). Возвращает экземпляр класса ShortCutListener.
     * @name y5.ShortCut.down
     * @memberOf y5.ShortCut
     * @function
     * @param {Array of Keys} keys Список сочетаний клавиш
     * @param {Function} callback Функция обработки нажатия клавиш
     * @param {Node} node Целевой элемент
     * @param {Object} options Параметры
     * @returns {Object} Экземпляр класса ShortCutListener
     *
     * @example
     * y5.ShortCut.down([{key: y5.ShortCut.ESC}], function() { alert('Esc'); }, document);
     */
    down: function(keys, callback, node, options) {
        return action(typeDown, listDown, keys, callback, node, options);
    },

    /**
     * Устанавливает обработчик клавиатурного сокращения на нажатие и отпускание клавиши (keypress). Возвращает экземпляр класса ShortCutListener.
     * @name y5.ShortCut.press
     * @memberOf y5.ShortCut
     * @function
     * @param {Array of Keys} keys Список сочетаний клавиш
     * @param {Function} callback Функция обработки нажатия клавиш
     * @param {Node} node Целевой элемент
     * @param {Object} options Параметры
     * @returns {Object} Экземпляр класса ShortCutListener
     *
     * @example
     * y5.ShortCut.press([{key: y5.ShortCut.ESC}], function() { alert('Esc'); }, document);
     */
    press: function(keys, callback, node, options) {
        return action(typePress, listPress, keys, callback, node, options);
    },

    /**
     * Удаляет обработчик клавиатурных сокращений. В случае успеха возвращает true.
     * @name y5.ShortCut.remove
     * @memberOf y5.ShortCut
     * @function
     * @param {Object} shortcut
     * @returns {Boolean}
     */
    remove: function(shortcut) {
        return this.removeDown(shortcut) || this.removePress(shortcut);
    },

    /**
     * Выключает обработку конкретного клавиатурного сокращения, которое было установлено на нажатие клавиши (keydown).
     * @name y5.ShortCut.removeDown
     * @memberOf y5.ShortCut
     * @function
     * @param {Object} shortcut
     * @returns {Boolean}
     */
    removeDown: function(shortcut) {
        return removeFromList(shortcut, listDown);
    },

    /**
     * Выключает обработку конкретного клавиатурного сокращения, которое было установлено на нажатие и отпускание клавиши (keypress).
     * @name y5.ShortCut.removePress
     * @memberOf y5.ShortCut
     * @function
     * @param {Object} shortcut
     * @returns {Boolean}
     */
    removePress: function(shortcut) {
        return removeFromList(shortcut, listPress);
    }
};

for (var name in KEYS) {
    y5.ShortCut[name] = KEYS[name];
}

var AEventListener = y5.AEventListener;

/// Init

var commonDownListener = keyDownListener;

if (y5.is_ie) {

    commonDownListener = function(e) {
        if (!e.repeat) {
            keyDownListener(e);
        }
        keyPressListener(e);
    };

} else if (y5.is_safari) {

    var keyUp, press = false;
    keyUp = new AEventListener('keyup', function() { press = false; keyUp.remove(); }, document, false);

    commonDownListener = function(e) {
        if (!press) {
            keyUp.add();
            keyDownListener(e);
        }
        press = true;

        keyPressListener(e);
    };

} else {

    new AEventListener('keypress', keyPressListener, document, true);

}

new AEventListener('keydown', commonDownListener, document, true);

y5.loaded('ShortCuts');

});