(function() {

var EMPTY = '',
    SPACE = ' ',
    trimRegexp = /(^[\s\xA0]+|[\s\xA0]+$)/g,
    isVoidRegexp = /^[\s\xA0]*$/,
    normalizeSpaceRegexp = /[\s\xA0]{2,}/g,
    escapeRegexpRe = /([\|\!\[\]\^\$\(\)\{\}\+\=\?\.\*\\])/g,
    stripTagsRegexp = /(<([^>]+)>)/ig,
    NLRegexp = /\r\n|\r|\n/g,
    wordsRegexp = /[^\s\xA0]+/g,
    escapeHTMLRe = /[&<>\"\']/g,
    escapeHTMLReplacer = function(c) {
        return '&#' + c.charCodeAt(0) + ';';
    },
    unescapeHTMLRe = /(&(lt|gt|quot|apos|amp|#\d+);|.)/gi,
    unescapeHTMLHash = {lt: '<', gt: '>', quot: '"', apos: "'", amp: '&'},
    unescapeHTMLReplacer = function(s, c, d) {
        return unescapeHTMLHash[d] || (d ? String.fromCharCode(d.substring(1)) : c);
    },
    // y5.Strings.dasherize
    dasherizeRegexp = /[A-Z]+[a-z]+/g,
    dasherizeFunction = function(s) {
            return '-' + s.toLowerCase()
    },
    // y5.Strings.camelize
    camelizeFunction = function(item, i) {
        if (i != 0) {
            return y5.Strings.capitalize(item);
        }
        return item;
    };

/**
 * Преобразует объект для y5.Strings.printf
 * @param {String} Формат
 * @param {Object} Данные для преобразования
 * @returns {String} Результирующая строка
 * @private
 */
function convertPrintf(format, data) {
    data = data.toString();

    var match = /^%(0?)(\d+)d$/.exec(format);
    if (match) {
        var padchar = match[1] || SPACE,
            padding = parseInt(match[2], 10) - data.length;
        return y5.Strings.repeat(padchar, padding) + data;
    }
    return data;
}

/**
 * @class Набор функций для работы со строками. Включает реализацию дополнительных (отсутствующих в стандарте JavaScript) функций, упрощающих работу со строковыми данными.
 * @static
 * @name y5.Strings
 */
y5.Strings = {
    /**
     * Проверяет строку на пустоту. Возвращает true, если строка пустая.
     * @name y5.Strings.isEmpty
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка для проверки
     * @returns {Boolean} Результат проверки
     */
    isEmpty: function(str) {
        return (str == EMPTY);
    },

    /**
     * Проверяет, что содержит строка - какие-либо значимые данные или только пробельные символы. Возвращает true, если строка состоит только из пробельных символов или является пустой.
     * @name y5.Strings.isVoid
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка для проверки
     * @returns {Boolean} Результат проверки
     * @example
     * ' \n\t' -> true
     */
    isVoid: function(str) {
        return (!str || isVoidRegexp.test(str));
    },

    /**
     * Проверяет, содержит ли строка заданную подстроку. Возвращает true, если вхождение заданной подстроки было найдено.
     * @name y5.Strings.contains
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @param {String} needle Строка для поиска
     * @param {Boolean} ci Поиск без учета регистра
     * @returns {Boolean} Результат проверки
     * @example
     * y5.Strings.contains("foobarbaz", "bar")
     * // -> true
     *
     * y5.Strings.contains("FooBarBaz", "bar")
     * // -> false
     *
     * y5.Strings.contains("FooBarBaz", "bar", true)
     * // -> true
     */
    contains: function(str, needle, ci) {
        if (ci) {
            str = str.toUpperCase();
            needle = needle.toUpperCase();
        }
        return str.indexOf(needle) !== -1;
    },

    /**
     * Проверяет, начинается ли строка с заданной подстроки. Возвращает true, если строка начинается с заданной подстроки.
     * @name y5.Strings.startsWith
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @param {String} needle Строка для поиска
     * @param {Boolean} ci Поиск без учета регистра
     * @returns {Boolean} Результат проверки
     * @example
     * y5.Strings.startsWith("foobar", "foo")
     * // -> true
     *
     * y5.Strings.startsWith("Foobar", "foo")
     * // -> false
     *
     * y5.Strings.startsWith("Foobar", "foo", true)
     * // -> true
     */
    startsWith: function(str, needle, ci) {
        if (ci) {
            str = str.toUpperCase();
            needle = needle.toUpperCase();
        }
        return str.indexOf(needle) === 0;
    },

    /**
     * Проверяет, оканчивается ли строка заданной подстрокой. Возвращает true, если строка заканчивается заданной подстрокой.
     * @name y5.Strings.endsWith
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @param {String} needle Строка для поиска
     * @param {Boolean} ci Поиск без учета регистра
     * @returns {Boolean} Результат проверки
     * @example
     * y5.Strings.endsWith("foobar", "bar")
     * // -> true
     *
     * y5.Strings.endsWith("FooBar", "bar")
     * // -> false
     *
     * y5.Strings.endsWith("FooBar", "bar", true)
     * // -> true
     */
    endsWith: function(str, needle, ci) {
        if (ci) {
            str = str.toUpperCase();
            needle = needle.toUpperCase();
        }
        return str.lastIndexOf(needle) + needle.length === str.length;
    },

    /**
     * Сравнивает две строки.
     * Возвращает отрицательное число, если str1  меньше, чем str2; положительное число, если str1 больше, чем str2, и 0 если строки равны.
     * @param {String} str1 Строка 1
     * @param {String} str2 Строка 2
     * @param {Boolean} ci Сравнение без учета регистра
     * @return {Number} Результат сравнения
     * @example
     * y5.Strings.compare('Style', 'styLe');
     * // -1
     *
     * y5.Strings.compare('style', 'StyLe', true);
     * // 0
     *
     * var fruits = ['banana', 'orange', 'lemon', 'grapefruit', 'cherry'];
     * fruits.sort(y5.Strings.compare).join(',');
     * // banana,cherry,grapefruit,lemon,orange
     */
    compare: function (str1, str2, ci) {
        if (ci) {
            str1 = str1.toLowerCase();
            str2 = str2.toLowerCase();
        }
        if (str1 == str2) {
            return 0;
        } else if (str1 < str2) {
            return -1;
        } else {
            return 1;
        }
    },

    /**
     * Удаляет начальные и конечные пробельные символы. Возвращает результирующую строку.
     * @name y5.Strings.trim
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Строка с удаленными пробелами
     * @example
     * "   foo  \n" -> "foo"
     */
    trim: function(str) {
        return str.replace(trimRegexp, EMPTY);
    },

    /**
     * Заменяет в строке множество пробельных символов на один пробел и удаляет начальные и конечные пробельные символы. Возвращает результат преобразования.
     * @name y5.Strings.normalizeSpace
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Преобразованная строка
     * @example
     * "   foo  \t  bar \n" -> "foo bar"
     */
    normalizeSpace: function(str) {
        return this.trim(str.replace(normalizeSpaceRegexp, SPACE));
    },

    /**
     * Экранирует все встреченные в строке символы, являющиеся специальными для регулярных выражений. Возвращает преобразованную строку.
     * Используется для вставки любой строки в регулярное выражение.
     * @name y5.Strings.escapeRegexp
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Преобразованная строка
     * @example
     * "{2}" -> "\{2\}"
     */
    escapeRegexp: function(text) {
        return text.replace(escapeRegexpRe, "\\$1");
    },

    /**
     * Определяет, какому символу соответствует заданный цифровой код. Возвращает символ по его коду.
     * @name y5.Strings.getCode
     * @memberOf y5.Strings
     * @function
     * @param {Number} code Код символа
     * @returns {String} Символ
     * @example
     * 65 -> "A"
     */
    getCode: function(code) {
        return String.fromCharCode(code);
    },

    /**
     * Подготавливает строку для безопасной вставки в HTML-код. Возвращает результат преобразования.
     * Все служебные HTML-символы в данной строке заменяются на их символьные аналоги. Данная функция часто используется для записи в innerHTML. Для обратного преобразования рекомендуется использовать функцию y5.Strings.unescapeHTML.
     * @name y5.Strings.escapeHTML
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Преобразованная строка
     * @example
     * "&lt;foo>" -> "&amp;lt;foo&amp;gt;"
     */
    escapeHTML: function(str) {
        return str.replace(escapeHTMLRe, escapeHTMLReplacer);
    },

    /**
     * Заменяет символьные аналоги служебных HTML-символов на реальные значения. Возвращает результат преобразования.
     * Функция, обратная escapeHTML.
     * @name y5.Strings.unescapeHTML
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Преобразованная строка
     * @example
     * "&amp;lt;foo&amp;gt;" -> "&lt;foo>"
     */
    unescapeHTML: function(str) {
        return str.replace(unescapeHTMLRe, unescapeHTMLReplacer);
    },

    /**
     * Выбирает из элемента DOM текстовое содержимое. Возвращает строку с полученным текстом.
     * @name y5.Strings.stripTags
     * @memberOf y5.Strings
     * @function
     * @param {String | HTMLElement} element Элемент
     * @returns {String} Текстовое содержимое элемента
     * @example
     * // &lt;div>&lt;i>foo&lt;/i> &lt;b>bar&lt;/b>&lt;/div>
     * y5.Strings.stripTags(div);
     * // -> "foo bar"
     *
     * y5.Strings.stripTags('&lt;i>foo&lt;/i> &lt;b>bar&lt;/b>');
     * // -> "foo bar"
     */
    stripTags: function(elem) {
        return (typeof elem == 'string' ? elem : elem.innerHTML).replace(stripTagsRegexp, EMPTY);
    },

    /**
     * Заменяет в строке все вхождения букв Ё и ё на Е и е соответственно. Возвращает результат преобразования.
     * @name y5.Strings.IoToIe
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Преобразованная строка
     * @example
     * "ёж" -> "еж"
     */
    IoToIe: function(str) {
        return str.replace(/[\u0451\u0401]/g, '\u0435');
    },

    /**
     * Подбирает любому числу правильную словоформу. Возвращает число и заданное слово в соответствующем данному числу склонении.
     * @name y5.Strings.plural
     * @memberOf y5.Strings
     * @function
     * @param {Number} number Число
     * @param {Array} forms Список форм слова (формы: 1, 2, 5, 0)
     * @param {Boolean} withoutNumber Преобразовать и вернуть только слово
     * @returns {String} Число с формой слова
     * @example
     * var forms = ['письмо', 'письма', 'писем', 'писем нет'];
     * y5.Strings.plural(0, forms);
     * // -> 'писем нет'
     *
     * y5.Strings.plural(1, forms);
     * // -> '1 письмо'
     *
     * y5.Strings.plural(2, forms);
     * // -> '2 письма'
     *
     * y5.Strings.plural(5, forms);
     * // -> '5 писем'
     *
     * y5.Strings.plural(5, forms, true);
     * // -> 'писем'
     *
     * var forms = ['письмо', 'письма', 'писем'];
     * y5.Strings.plural(0, forms);
     * // -> '0 писем'
     *
     * @todo сделать формат вывода
     */
    plural: function(num, forms, withoutNumber) {
        var res = 2;
        var num_10 = num % 10;
        var num_100 = num % 100;

        if (num == 0) {
            // если задана форма для нулевого значения, то выводим только его
            if (forms[3]) {
                return forms[3];
            }
        } else {
            if (num_100 < 5 || num_100 > 20) {
                if (num_10 == 1) {
                    res = 0;
                } else {
                    if (num_10 >= 2 && num_10 <= 4) {
                        res = 1;
                    }
                }
            }
        }

        if (withoutNumber) {
            return forms[res];            
        } else {
            return num + SPACE + forms[res];
        }
    },

    /**
     * Позволяет выводить число с разными формами слова.
     * @param {Number} number Число
     * @param {Array} forms Список форм слова (формы: 1, 5, 2, 0)
     * @returns {String} Число с формой слова
     * @deprecated y5.Strings.plural
     */
    conversion: function(num, words) {
        return this.plural(num, [words[0], words[2], words[1], words[3]]);
    },

    /**
     * Переводит первый символ строки в верхний регистр. Возвращает преобразованную строку.
     * @name y5.Strings.capitalize
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Преобразованная строка
     * @example
     * "foo" -> "Foo"
     */
    capitalize: function(str) {
        return str.charAt(0).toUpperCase() + str.substr(1).toLowerCase();
    },

    /**
     * Переводит слова, разделенные дефисами, в форму camelCase. Возвращает результат преобразования.
     * @name y5.Strings.camelize
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Преобразованная строка
     * @example
     * "foo-bar-baz" -> "fooBarBaz"
     */
    camelize: function(str) {
        return str.split('-').map(camelizeFunction).join(EMPTY);
    },

    /**
     * Разделяет дефисами слова в форме camelCase. Возвращает результат преобразования.
     * @name y5.Strings.dasherize
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} Преобразованная строка
     * @example
     * y5.Strings.dasherize("fooBarBaz");
     * // "foo-bar-baz"
     */
    dasherize: function (str) {
        return str.replace(dasherizeRegexp, dasherizeFunction);
    },

    /**
     * Повторяет строку заданное число раз. Возвращает строку, состоящую из заданного количества повторений исходной строки.
     * @name y5.Strings.repeat
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @param {Number} count Число повторений
     * @returns {String} Строка, продублированная number раз
     * @example
     * y5.Strings.repeat("foobar", 3);
     * // -> "foobarfoobarfoobar"
     *
     * y5.Strings.repeat("foo", 0);
     * // -> ""
     */
    repeat: function(str, count) {
        if (count < 1) {
            return EMPTY;
        }

        return (new Array(count + 1)).join(str);
    },

    /**
     * Заменяет все символы перевода в строке на теги &lt;br&gt;. Возвращает преобразованную строку.
     * @name y5.Strings.nl2br
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @param {Boolean} xhtml XHTML-совместимый вывод. true -> "&lt;br />", иначе - "&lt;br>"
     * @returns {String} HTML-строка
     * @example
     * "foo\nbar" -> "foo&lt;br>bar"
     */
    nl2br: function(str, xhtml) {
        return str.replace(NLRegexp, xhtml ? '<br />' : '<br>');
    },

    /**
     * Преобразовывает текст в HTML (экранирует все HYML-символы и заменяет перевод строк на &lt;br&gt;)
     * @name y5.Strings.text2html
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {String} HTML-строка
     */
    text2html: function(str) {
        return this.nl2br(this.escapeHTML(str))
    },

    /**
     * Преобразует строку в массив слов. Возвращает получившийся массив.
     * @name y5.Strings.words
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {Array} Массив слов
     * @example
     * y5.Strings.words('foo bar');
     * // -> ['foo', 'bar']
     *
     * y5.Strings.words('');
     * // -> []
     */
    words: function(str) {
        return str.match(wordsRegexp) || [];
    },

    /**
     * Подсчитывает количество слов в строке. Возвращает полученное число.
     * @name y5.Strings.wordsCount
     * @memberOf y5.Strings
     * @function
     * @param {String} string Строка
     * @returns {Number} Число слов
     * @example
     * y5.Strings.wordsCount('foo bar');
     * // -> 2
     */
    wordsCount: function(str) {
        return this.words(str).length;
    },

    /**
     * Реализует поддержку форматированного вывода данных. Формат соответствует популярному формату функции printf языка программирования C. Возвращает отформатированную по заданному шаблону строку.
     * @name y5.Strings.printf
     * @memberOf y5.Strings
     * @function
     * @param {String} format Строка формата вывода
     * @param {Object | Array} arg1 аргумент 1 (если массив, то он используется в качестве списка аргументов)
     * @param {Object} arg2 аргумент 2
     * @param {Object} argN аргумент n
     * @returns {String} Результат
     */
    printf: function(str, data) {
        var values = data;
        var args = arguments, length = args.length;
        if (length > 2) {
            values = [];
            for (var i = 1; i < length; i++) {
                values.push(args[i]);
            }
        } else if (typeof data != 'object') {
            values = [data];
        }

        var k = 0;
        function replacer(token) {
            var value = values[k]; k++;
            return convertPrintf(token, typeof value != y5.UNDEF ? value : EMPTY);
        }

        return str.replace(/%(s|\d*d)/g, replacer).replace(/%%/g, '%');
    },

    /**
     * Создает пустую строку. Возвращает строку с нулевой длиной.
     * @name y5.Strings.EMPTY
     * @constant
     * @type String
     */
    EMPTY: EMPTY,

    /**
     * Создает строку, состоящую из одного пробела. Возвращает полученную строку.
     * @name y5.Strings.SPACE
     * @constant
     * @type String
     */
    SPACE: SPACE,

    /**
     * Создает строку, состоящую из одного неразрывного пробела. Возвращает полученную строку.
     * @name y5.Strings.NBSP
     * @constant
     * @type String
     */
    NBSP: '\u00A0'
};

var Strings = y5.Strings;

/// aliases

/**
 * @name y5.Strings.strip
 * @memberOf y5.Strings
 * @see y5.Strings.trim
 * @function
 */
Strings.strip = Strings.trim;

/**
 * @name y5.Strings.times
 * @memberOf y5.Strings
 * @see y5.Strings.repeat
 * @function
 */
Strings.times = Strings.repeat;

/**
 * @name y5.Strings.isBlank
 * @memberOf y5.Strings
 * @see y5.Strings.isVoid
 * @function
 */
Strings.isBlank = Strings.isVoid;

/**
 * @name y5.Strings.normalize
 * @memberOf y5.Strings
 * @see y5.Strings.normalizeSpace
 * @function
 */
Strings.normalize = Strings.normalizeSpace;

/**
 * @name stripHTML
 * @memberOf y5.Strings
 * @function
 * @deprecated y5.Strings.stripTags
 */
Strings.stripHTML = Strings.stripTags;

})();

y5.loaded('Strings');
