y5.require('Utils', 'Widget.Menu', 'EventsExt', 'Fx', 'Elements', function() {

    var Notify = y5.Notify,
        Events = y5.Events,
        Types = y5.Types,
        Widget = y5.Widget;

    /**
     * Класс-обработчик полей автозаполенния.
     * Объект finder должен предоставлять метод process, который должен возвращать объект cо свойствами elements и tokens.
     * elements представляет собой массив из объектов {id:123, html:'text found in html'|label:'plain text'}. Формат соответствует данным о пунктах меню для Widget.Menu.
     * tokens - значение, которое необходимо подсвечивать; может представлять собой строку, массив или регулярное выражение.
     *
     * Объект генерирует на input следующие события:
     *   * y5:Suggest:open - при появлении поля автозаполения
     *   * y5:Suggest:close - при скрытии поля автозаполения
     *   * y5:Suggest:select - при выборе значения из списка. Callback-функция вызывается с аргументами (id, value)
     * 
     * @class
     * @constructor
     * @name y5.Suggest
     * @param {Element} input Поле
     * @param {Object} finder Объект, производящий поиск результатов для автозаполнения
     *
     * @example
     * var Finder = {
     *     process: function (value) {
     *         return {
     *             elements: [{id:1,html:'result 1'},{id:2,html:'result 2'},{id:3,html:'result 3'}],
     *             tokens: value
     *         }
     *     }
     * }
     *
     * var input = y5.$('input4AutoComplete');
     * new y5.Suggest(input, Finder);
     */
    y5.Suggest = function (input, finder) {
        this.input = input;
        input.setAttribute('autocomplete', 'off');

        this.finder = finder;
        this.popup = y5.Utils.generateUniqueId();
        this.result = null;

        this.menuListener = (function(context) {
            return function() {
                context.processSelect.apply(context, arguments)
            }
        })(this);

        this.inputObserver = new y5.InputObserver(this.show, input, true, this);

        Events.observe('y5:esc', this.esc, this.popup, true, this);
        Events.observe('y5:activate', this.activate, this.popup, true, this);

        y5.Events.observe('y5:init', function(widget) {
            y5.Elements.css(widget.window, 'opacity', 0);
            this.effect = y5.Fx.Morph(widget.window, {opacity: 1}, {duration: .1});
        }, this.popup, true, this);

        y5.Events.observe('y5:make', function(widget){
            y5.Elements.css(widget.window, 'width', input.offsetWidth);
            this.effect.start();
        }, this.popup, true, this);

        y5.Events.observe('y5:kill', function(){
            this.effect.reset();
        }, this.popup, true, this);
    }

    y5.Suggest.prototype = {
        show: function(el) {
            this.result = this.finder.process(this.input.value);

            if (this.result.elements.length > 0) {
                Widget.make(this.popup, 'Widget.Menu', {element: el, saveState: false});
                Notify('y5:Suggest:open', this.input);
            } else {
                this.closeWidget();
                Notify('y5:Suggest:noresult', this.input);
            }
        },

        esc: function() {
            this.input.focus();
            return true;
        },

        closeWidget: function () {
            Widget.kill(this.popup);
            Notify('y5:Suggest:close', this.input);
        },

        processSelect: function (widget, e) {
            e.preventDefault();
            this.select(y5.Dom.textContent(y5.Dom.getChildren(widget.container, 'div', 'y5-w-menu-item')[widget.elements[widget.activeItem]]), widget.activeItem);
        },

        select: function (value, id) {
            this.input.value = value;
            // сохраняем состояние для inputObserver, чтобы не вызывался обработчик
            this.inputObserver.setInitState();
            this.closeWidget();

            Notify('y5:Suggest:select', this.input, [id, value]);
        },

        activate: function() {
            var results = this.result.elements;

            // если результат 1, и он равен значению в input, то заменяем значение в input этим результатом
            if (results.length == 1 && results[0].html.toLowerCase() == this.input.value.toLowerCase()) {
                this.select(this.result.elements[0].html, this.result.elements[0].id);

            } else {
                var items = [],
                    item,
                    length = results.length,
                    i=0,
                    tokens = this.result.tokens,
                    doReplace = false;

                if (Types.def(tokens)) {
                    if (Types.string(tokens)) {
                        tokens = new RegExp(tokens, 'gi');

                    } else if (Types.array(tokens)) {
                        tokens = new RegExp(tokens.join('|'), 'gi');

                    } else if (!Types.regexp(tokens)) {
                        tokens = null;
                    }
                }

                doReplace = Types.regexp(tokens);

                for (; i<length; i++) {
                    item = results[i];
                    if (doReplace && item.html) {
                        item.html = item.html.replace(tokens, '<strong>$&</strong>');                       
                    }
                    item.listener = this.menuListener
                    items[i] = item;
                }

                // устанавливаем пункты меню
                Notify('y5:setItems', this.popup, items);
            }
            this.input.focus();
        }
    };

    y5.loaded('Suggest');
});