(function() {

var UNDEF = y5.UNDEF;

function checkDomNode(obj, type) {
    return (obj && obj.nodeType && obj.nodeType == type) || false;
}

/**
 * @class Модуль для работы с типами объектов. Содержит определения констант, соответствующих разным типам объектов, и набор функций для выяснения типа объекта.
 * @name y5.Types
 * @static
 */
y5.Types = {
    /// consts

    // standard

    /**
     * Константа, обозначающая тип «неопределенный». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.UNDEF
     * @memberOf y5.Types
     * @constant
     * @link y5.Types.UNDEFINED
     * @type Number
     */
    UNDEF:          1 << 0,

    /**
     * Константа, обозначающая тип «неопределенный». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.UNDEFINED
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    UNDEFINED:      1 << 0,

    /**
     * Константа, обозначающая тип «объект». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.OBJECT
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    OBJECT:         1 << 1,

    /**
     * Константа, обозначающая тип «функция». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.FUNCTION
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    FUNCTION:       1 << 2,

    /**
     * Константа, обозначающая тип «число». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.NUMBER
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    NUMBER:         1 << 3,

    /**
     * Константа, обозначающая тип «строка». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.STRING
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    STRING:         1 << 4,

    /**
     * Константа, обозначающая «логический» тип. Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.BOOLEAN
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    BOOLEAN:        1 << 5,

    // user
    /**
     * Константа, обозначающая тип «дата». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.DATE
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    DATE:           1 << 10,

    /**
     * Константа, обозначающая тип «регулярное выражение». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.REGEXP
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    REGEXP:         1 << 11,

    /**
     * Константа, обозначающая тип «массив». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.ARRAY
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    ARRAY:          1 << 12,

    /**
     * Тип null.
     * @name y5.Types.NULL
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    NULL:           1 << 13,

    /**
     * Константа, обозначающая тип «событие». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.EVENT
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    EVENT:          1 << 14,

    /**
     * Константа, обозначающая тип «узел документа». Возвращает цифровой код, соответствующий данному типу.
     * @name y5.Types.NODE
     * @memberOf y5.Types
     * @constant
     * @type Number
     */
    NODE:           1 << 15,

    TYPES: {
        'undefined': 1 << 0,
        'object':    1 << 1,
        'function':  1 << 2,
        'number':    1 << 3,
        'string':    1 << 4,
        'boolean':   1 << 5
    },

    /// methods

    /**
     * Возвращает числовой код типа объекта – одну из констант y5.Types.
     * @name y5.Types.type
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект
     * @returns {Number} Константа y5.Types
     *
     * @example
     * y5.Types.type(1);
     * // -> y5.Types.NUMBER
     *
     * y5.Types.type(new Date);
     * // -> y5.Types.DATE
     *
     * y5.Types.type([1, 2, 3]);
     * // -> y5.Types.ARRAY
     *
     * y5.Types.type(document);
     * // -> y5.Types.NODE
     */
    type: function(obj) {
        var type = this.TYPES[typeof obj];

        // NULL
        if (obj === null) {
            return this.NULL;
        }

        // DOM NODE
        if (type == this.OBJECT) {
            if (obj.nodeName || this.document(obj) /* IE 5.0 */) {
                return this.NODE;
            }
        }

        // ARRAY, REGEXP, DATE
        if (type == this.OBJECT || type == this.FUNCTION) {
            switch(obj.constructor) {
                case Array:
                    return this.ARRAY;

                case RegExp:
                    return this.REGEXP;

                case Date:
                    return this.DATE;
            }
        }

        // EVENT
        if (this.event(obj)) {
            return this.EVENT;
        }

        return type;
    },

    /**
     * Проверяет, является ли тип объекта одним из типов, заданных параметром mask. В случае успеха проверки возвращает true.
     * @name y5.Types.test
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @param {Number} mask Маска для проверки (задается в виде TYPE1 | TYPE2 | ...)
     * @returns {Boolean} Результат проверки
     *
     * @example
     * // проверить, что тип объекта Date или RegExp
     * y5.Types.test(obj, y5.Types.DATE | y5.Types.REGEXP);
     */
    test: function(obj, mask) {
        return !!(this.type(obj) & mask);
    },

    // standard

    /**
     * Проверяет, является ли объект определенным. В случае успеха проверки возвращает true.
     * @name y5.Types.def
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    def: function(obj) {
        return typeof obj != UNDEF;
    },

    /**
     * Проверяет, является ли объект неопределенным. В случае успеха проверки возвращает true.
     * @name y5.Types.undef
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    undef: function(obj) {
        return typeof obj == UNDEF;
    },

    /**
     * Проверяет, принадлежит ли объект к типу Object. В случае успеха проверки возвращает true.
     * @name y5.Types.object
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    object: function(obj) {
        return typeof obj == 'object';
    },

    /**
     * Проверяет, является ли объект функцией. В случае успеха проверки возвращает true.
     * @name y5.Types.func
     * @memberOf y5.Types
     * @function
     * @param {Object} object Рбъект для проверки
     * @returns {Boolean} Результат проверки
     */
    func: function(obj) {
        return typeof obj == 'function';
    },

    /**
     * Проверяет, является ли объект числом. В случае успеха проверки возвращает true.
     * @name y5.Types.number
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    number: function(obj) {
        return typeof obj == 'number';
    },

    /**
     * Проверяет, является ли объект строкой. В случае успеха проверки возвращает true.
     * @name y5.Types.string
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    string: function(obj) {
        return typeof obj == 'string';
    },

    /**
     * Проверяет, принадлежит ли объект к типу Boolean. В случае успеха проверки возвращает true.
     * @name y5.Types.bool
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    bool: function(obj) {
        return typeof obj == 'boolean';
    },

    /**
     * Проверяет, является ли объект null. В случае успеха проверки возвращает true.
     * @name y5.Types.nul
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    nul: function(obj) {
        return obj === null;
    },

    // user

    /**
     * Проверяет, является ли объект массивом. В случае успеха проверки возвращает true.
     * @name y5.Types.array
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    array: function(obj) {
        return obj instanceof Array;
    },

    /**
     * Проверяет, является ли объект регулярным выражением. В случае успеха проверки возвращает true.
     * @name y5.Types.regexp
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    regexp: function(obj) {
        return obj instanceof RegExp;
    },

    /**
     * Проверяет, является ли объект датой. В случае успеха проверки возвращает true.
     * @name y5.Types.date
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    date: function(obj) {
        return obj instanceof Date;
    },

    /**
     * Проверяет, является ли объект событием. В случае успеха проверки возвращает true.
     * @name y5.Types.event
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    event: function(obj) {
        // Правильно: obj instanceof Event
        // Для IE
        return obj && typeof obj.type != UNDEF && typeof (obj.stopPropagation || obj.cancelBubble) != UNDEF;
    },

    // dom

    /**
     * Проверяет, является ли объект элементом DOM. В случае успеха проверки возвращает true.
     * @name y5.Types.element
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    element: function(obj) {
        return checkDomNode(obj, Node.ELEMENT_NODE);
    },

    /**
     * Проверяет, является ли объект атрибутом DOM. В случае успеха проверки возвращает true.
     * @name y5.Types.attribute
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    attribute: function(obj) {
        return checkDomNode(obj, Node.ATTRIBUTE_NODE);
    },

    /**
     * Проверяет, является ли объект текстовым узлом DOM. В случае успеха проверки возвращает true.
     * @name y5.Types.text
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    text: function(obj) {
        return checkDomNode(obj, Node.TEXT_NODE);
    },

    /**
     * Проверяет, является ли объект документом. В случае успеха проверки возвращает true.
     * @name y5.Types.document
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    document: function(obj) {
        // для старых браузеров IE 5.0
        return (obj && typeof obj.documentElement != UNDEF) || false;

        // правильно
        // return checkDomNode(obj, Node.DOCUMENT_NODE);
    },

    /**
     * Проверяет, является ли объект узлом комментария. В случае успеха проверки возвращает true.
     * @name y5.Types.comment
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    comment: function(obj) {
        return checkDomNode(obj, Node.COMMENT_NODE);
    },

    /**
     * Проверяет, является ли объект узлом DOM. В случае успеха проверки возвращает true.
     * @name y5.Types.node
     * @memberOf y5.Types
     * @function
     * @param {Object} object Объект для проверки
     * @returns {Boolean} Результат проверки
     */
    node: function(obj) {
        return (obj && typeof obj.nodeType != UNDEF);
    }
};

})();