(function() {

var Types = y5.Types,
    NOT_OBJECT = Types.NODE | Types.BOOLEAN | Types.NUMBER | Types.STRING | Types.EVENT | Types.REGEXP | Types.FUNCTION | Types.NULL;

/**
 * @class Набор полезных функций-утилит общего назначения.
 * @static
 * @name y5.Utils
 */
y5.Utils = {
    /**
     * Счетчик генерируемых ID.
     * @name y5.Utils.counterId
     * @memberOf y5.Utils
     * @type Number
     * @private
     */
    counterId: 0,

    /**
     * Генерирует случайный ID. Возвращает результат в виде строки символов.
     * @name y5.Utils.generateId
     * @memberOf y5.Utils
     * @function
     * @param {String} [prefix] Префикс (по умолчанию = '')
     * @returns {String} Случайный ID
     */
    generateId: function(prefix) {
        return (prefix || '') + ((new Date()).getTime() + Math.round(Math.random() * 10000));
    },

    /**
     * Генерирует уникальный ID. Возвращает результат в виде строки символов.
     * @name y5.Utils.generateUniqueId
     * @memberOf y5.Utils
     * @function
     * @returns {String} Уникальный ID
     *
     * @example
     * y5.Utils.generateUniqueId();
     * // -> 'y5__id45'
     */
    generateUniqueId: function() {
        return 'y5__id' + (++this.counterId);
    },

    /**
     * Генерирует уникальный ID и присваивает его свойству uniqueID переданного объекта.
     * Если свойство uniqueID объекта существует, то значение этого свойства не изменяется. Возвращает значение uniqueID в виде строки символов.
     * @name y5.Utils.getUniqueId
     * @memberOf y5.Utils
     * @function
     * @param {Object} object Объект
     * @returns {String} Уникальный ID
     */
    getUniqueId: function(object) {
        if (object === document) {
            return this.documentID;
        }
        return object.uniqueID || this.setUniqueId(object);
    },

    /**
     * Присваивает свойству uniqueID объекта переданное значение. Если параметр uniqueID не определен, то генерируется новый уникальный идентификатор и полученное значение присваивается свойству uniqueID объекта. Возвращает значение свойства uniqueID в виде строки символов.
     * @name y5.Utils.setUniqueId
     * @memberOf y5.Utils
     * @function
     * @param {Object} object Объект
     * @param {String} [uniqueID] Уникальный ID (если неопределен, то генерируется новый)
     * @returns {String} Уникальный ID
     *
     * @example
     * y5.Utils.setUniqueId(object, 'foo');
     * // -> 'foo'
     *
     * y5.Utils.setUniqueId(object);
     * // -> 'y5__id46'
     */
    setUniqueId: function(object, uniqueID) {
        if (typeof uniqueID == y5.UNDEF) {
            uniqueID = this.generateUniqueId();
        }
        return (object.uniqueID = uniqueID);
    },

    /**
     * Сравнивает два объекта. Возвращает true, если оба объекта на самом деле - это один и тот же объект.
     * @name y5.Utils.isEqual
     * @memberOf y5.Utils
     * @function
     * @param {Object} first Объект
     * @param {Object} second Объект
     * @returns {Boolean} Результат сравнения
     */
    isEqual: function(first, second) {
        return this.getUniqueId(first) == this.getUniqueId(second);
    },

    /**
     * Решает проблему браузера IE, связанную с некорректной отрисовкой тега select и объекта Flash, в результате которой эти визуальные элементы перекрывают любые другие элементы страницы, даже находящиеся на  слоях выше select или Flash. Проблема решается путем создания элемента iframe и «подкладывания» его под элемент (обычно div), который необходимо корректно отобразить поверх select или Flash.
     * @name y5.Utils.fakeFrame
     * @memberOf y5.Utils
     * @type Object
     */
    fakeFrame: {
        frame: null,
        transparentFrame: null,

        init: function (obj) {
            if (!y5.is_ie7down) {
                return;
            }
            this.frame = y5.Dom.$('fakeFrame');
            if (!this.frame) {
                this.create(obj);
            }

            return [this.frame, this.transparentFrame];
        },

        create: function (obj) {
            this.transparentFrame = document.createElement('<iframe id="fakeTransparentFrame" src="about:blank" frameborder="0" tabindex="-1" style="filter:Alpha(opacity=1); position: absolute;">');
            this.frame = document.createElement('<iframe id="fakeFrame" src="about:blank" frameborder="0" tabindex="-1" allowtransparency="true" style="FILTER: chroma(color=#FFFFFF); position: absolute;">');
            this.transparentFrame.style.zIndex = y5.Elements.css(obj, 'z-index') - 2;
            this.frame.style.zIndex = y5.Elements.css(obj, 'z-index') - 1;
            this.frame.style.display = this.transparentFrame.style.display = 'none';
            y5.Dom.getBody().appendChild(this.transparentFrame);
            y5.Dom.getBody().appendChild(this.frame);
        },

        adjust: function (obj) {
            if (!this.frame) {
                return;
            }

            if (obj.offsetHeight > 0) {
                if (this.transparentFrame.style.zIndex == -2) {
                    this.transparentFrame.style.zIndex = y5.Elements.css(obj, 'z-index') - 2;
                    this.frame.style.zIndex = y5.Elements.css(obj, 'z-index') - 1;
                }
                this.frame.style.width = this.transparentFrame.style.width = obj.offsetWidth + 'px';
                this.frame.style.height = this.transparentFrame.style.height = obj.offsetHeight + 'px';
                this.frame.style.left = this.transparentFrame.style.left = obj.offsetLeft + 'px';
                this.frame.style.top = this.transparentFrame.style.top = obj.offsetTop + 'px';
            }
            this.frame.style.display = this.transparentFrame.style.display = obj.style.display;
        }
    },

    /**
     * Массив символов шестнадцатиричной системы счисления.
     * @private
     * @name y5.Utils.hexDigit
     * @memberOf y5.Utils
     * @type Array
     */
    hexDigit: '0123456789ABCDEF'.split(''),

    /**
     * Преобразует десятичное число в двузначное шестнадцатеричное. Возвращает полученное значение в виде строки.
     * @name y5.Utils.dec2hex
     * @memberOf y5.Utils
     * @function
     * @param {Number} number Число
     * @returns {String} 16-тиричное число
     *
     * @example
     * y5.Utils.dec2hex(10);
     * // -> '0A'
     */
    dec2hex: function(n) {
        return this.hexDigit[n >> 4] + this.hexDigit[n & 15];
    },

    /**
     * Преобразует шестнадцатеричное число в десятичное. Возвращает результат преобразования.
     * @name y5.Utils.hex2dec
     * @memberOf y5.Utils
     * @function
     * @param {String} number 16-тиричное число
     * @returns {Number} Десятичное число
     *
     * @example
     * y5.Utils.hex2dec('A');
     * // -> 10
     */
    hex2dec: function(n) {
        return parseInt(n, 16);
    },

    /**
     * Поочередно копирует свойства нескольких объектов в новый. Возвращает объект, получившийся в результате слияния исходных.
     * @name y5.Utils.objectCopy
     * @memberOf y5.Utils
     * @function
     * @param {Object} object1 Первый объект
     * @param {Object} object2 Второй объект
     * @param {Object} objectN n-объект
     * @returns {Object} Объект, получившийся в результате слияния объектов
     *
     * @example
     * y5.Utils.objectCopy({a: 1, b: 3}, {b: 2, c: 3});
     * // -> {a: 1, b: 2, c: 3}
     */
    objectCopy: function(destination) {
        var key,
            result = {},
            argsCnt = arguments.length;

        for (key in destination) {
            result[key] = destination[key];
        }

        if (argsCnt == 1) {
            return result;

        } else if (argsCnt == 2) {
            var source = arguments[1],
                src,
                data;
            for (key in source) {
                src = source[key];
                data = {};

                if (Types.test(src, NOT_OBJECT)) {
                    data = src;
                // Date
                } else if (Types.date(src)) {
                    data = new Date(src);
                // Array
                } else if (Types.array(src)) {
                    data = [].concat(src);
                // Object
                } else {
                    if (Types.def(result[key])) {
                        data = result[key];
                    }
                    data = this.objectCopy(data, src);
                }

                result[key] = data;
            }

        } else {
            key=1;
            for (; key<argsCnt; key++) {
                result = this.objectCopy(result, arguments[key]);
            }
        }
        return result;
    },

    /**
     * Расширяет функции прототипа первого объекта функциями прототипа второго объекта.
     * @name y5.Utils.objectExtends
     * @memberOf y5.Utils
     * @function
     * @param {Object} destination Расширяемый объект
     * @param {Object} source Объект с функциями расширения
     * @param {Object} sourceName Имя объекта расширения
     */
    objectExtends: function(destination, source, source_name) {
        source_name = source_name || source.toString().match(/function\s*([^\(]+)\(/)[1];

        var key,
            destProto = destination.prototype,
            sourceProto = source.prototype;

        destProto[source_name] = source;

        for (key in sourceProto) {
            // если такого свойства нет в destination, то копируем его из source
            if (Types.undef(destProto[key])) {
                destProto[key] = sourceProto[key];

            // если свойство в destination и в source - объекты, то объединяем их  через objectCopy    
            } else if (Types.object(destProto[key]) && Types.object(sourceProto[key])) {
                destProto[key] = this.objectCopy(sourceProto[key], destProto[key]);
            }
            /*
             // если свойство есть, то ничего не делаем
             } else {}
            */
        }
    },

    /**
     * Запускает выполнение метода указанного объекта с периодичностью, заданной в параметре timeout. Возвращает уникальный ID таймера.
     * @name y5.Utils.setTimeout
     * @memberOf y5.Utils
     * @function
     * @param {Function} method Метод
     * @param {Number} timeout Таймаут (мсек)
     * @param {Object} [context] Объект, в контексте которого запускается функция (по умолчанию null)
     * @param {Object} [arg1] Аргумент вызываемой функции
     * @param {Object} [arg2]
     * @param {Object} [...]
     * @param {Object} [argN]
     * @returns {ID} ID таймера
     */
    setTimeout: function(func, timeout, context /* args */) {
        var args = [];

        for (var i = 3, l = arguments.length; i < l; i++) {
            args.push(arguments[i]);
        }

        function call() {
            func.apply(context, args);
        }

        return window.setTimeout(call, timeout);
    },

    /**
     * Форматирует число в строку с разделителем
     * @name y5.Utils.formatNumber
     * @memberOf y5.Utils
     * @function
     * @param {Number|String} num Число
     * @param {String} [exponent_delimiter] Разделитель порядков. По умолчанию: пробел
     * @param {String} [float_delimiter] Разделитель дробной части. По умолчанию: точка
     * @returns {String} Отформатированное число
     * @example
     * y5.Utils.formatNumber('10000', ' ');
     * // -> '10 000'
     *
     * y5.Utils.formatNumber('-10000.02', ',', '.');
     * // -> '-10,000.02'
     *
     */
    formatNumber: function (num, exponent_delimiter, float_delimiter) {
        num = parseFloat(num.toString(), 10);
        if (isNaN(num)) {
            return;
        }

        exponent_delimiter = Types.string(exponent_delimiter) ? exponent_delimiter : ' ';
        float_delimiter = float_delimiter ? float_delimiter : '.';

        // определяем знак
        var sgn = num < 0 ? '-' : '',
        // отделяем дробную часть
            strNum = num.toString(),
            pointIndex = strNum.indexOf('.'),
            floatPart = 0;
        if (pointIndex != -1) {
            floatPart = strNum.substr(pointIndex + 1);
        }

        // берем число без знака и отсекаем дробную часть
        num = Math.floor(Math.abs(num)).toString();

        var startIndex = num.length % 3,
            result = num.substr(0, startIndex),
            cnt = Math.floor(num.length / 3),
            i = 0;
        for (; i < cnt; i++) {
            result += exponent_delimiter + num.substr(3 * i + startIndex, 3);
        }

        // если количество цифр кратно 3, то убираем первый exponent_delimiter
        if (startIndex == 0) {
            result = result.substr(1);
        }

        if (floatPart) {
            result += float_delimiter + floatPart;
        }

        return sgn + result;
    }
};

y5.Utils.documentID = y5.Utils.generateId('y5__');

/**
 * @name y5.Utils.getUniqueID
 * @memberOf y5.Utils
 * @function
 * @deprecated y5.Utils.getUniqueId
 */
y5.Utils.getUniqueID = y5.Utils.getUniqueId;

y5.loaded('Utils');

})();