/**
 * @name XmlRpc 
 * @fileOverview
 * Работа с XML-RPC протоколом (http://www.xmlrpc.com/).
 * Позволяет формулировать вызовы в терминах JavaScript и преобразует XML-RPC-ответы в JavaScript-объекты.
 * Из-за особенностей транспорта (в y5.Request.XML не работают кросс-доменные запросы) может использоваться только на домене XML-RPC-сервиса.
 */

y5.require(['Strings', 'Request.XML'], function(){

/**
 * Абстрактный класс для создания XML-RPC-запросов.
 *
 * @name y5.XmlRpc
 * @class
 * @constructor
 * @param {Object} params Объект параметров, поддерживаются:
 *   {String} url Путь к XML-RPC-сервису
 *
 * @example
 * var xmlrpc = new y5.XmlRpc();
 * xmlrpc.call({
 *     method: 'login',
 *     params: ['mylogin', 'mypassword']
 *     onload: function(res) { alert(res) },
 *     onerror: function(e) { alert(e) }
 * });
 */
y5.XmlRpc = function(params) {
    this.init(params || {});
};

y5.XmlRpc.prototype = {
    /**
     * Инициализация.
     * @private
     * @param {Object} params Объект параметров, поддерживаются:
     *   {String} url Путь к XML-RPC-сервису
     */
    init: function(params) {
        this.url = params.url || '/rpc/xmlrpc';
    },

    /**
     * Переводит XML-дерево XML-RPC-ответа в JavaScript-объекты.
     * @param {Object} xml XML-дерево (НЕ строка)
     * @returns {Object} JavaScript-объект результата
     * @private
     */
    xmlrpc2json: function(xml) {
        function isEmpty(o) {
            if (!o) return true;
            if (y5.Types.array(o) && o.length == 0) return true;
            if (y5.Types.object(o)) {
                for (var i in o) {
                    if ((o.hasOwnProperty && o.hasOwnProperty(i)) || o[i]) return false;
                }
                return true;
            }
            return false;
        }

        function value2json(valueNode) {
            switch (valueNode.firstChild.nodeName) {
                case '#text':
                    return valueNode.firstChild.nodeValue;

                case 'i4':
                case 'int':
                case 'double':
                    return Number(valueNode.firstChild.firstChild.nodeValue);

                case 'boolean':
                    return valueNode.firstChild.firstChild.nodeValue == 1;

                case 'dateTime.iso8601':
                case 'base64':
                case 'string':
                    return valueNode.firstChild.firstChild.nodeValue;

                case 'array':
                    var array = [];
                    for (var n=valueNode.firstChild.firstChild.firstChild; n; n=n.nextSibling) {
                        var json = value2json(n);
                        if (json) array[array.length] = json;
                    }
                    if (isEmpty(array)) return null;
                    return array;

                case 'struct':
                    var struct = {};
                    for (var n=valueNode.firstChild.firstChild; n; n=n.nextSibling) {
                        var json = value2json(n.childNodes[1]);
                        if (json) struct[n.firstChild.nodeValue || n.firstChild.firstChild.nodeValue] = json;
                    }
                    if (isEmpty(struct)) return null;
                    return struct;

                default:
                    return null;
            }
        }

        var result = {fault: {}};
        if (xml.firstChild.nodeName == 'methodResponse') {
            var methodResult = xml.firstChild.firstChild;
            if (methodResult.nodeName == 'params') {
                var json = value2json(methodResult.firstChild.firstChild);
                if (y5.Types.type(json) != y5.Types.NULL) result = {param: json};
            }
            if (methodResult.nodeName == 'fault') {
                var json = value2json(methodResult.firstChild);
                if (y5.Types.type(json) != y5.Types.NULL) result['fault'] = json;
            }
        }
        return result;
    },

    /**
     * Переводит массив JavaScript-объектов в XML-RPC-параметры.
     * @param {Array} json Массив JavaScript-объектов
     * @returns {String} XML-строка, содержащая параметры XML-RPC
     * @private
     */
    json2xmlrpc: function(json) {
        function json2value(json) {
            switch (y5.Types.type(json)) {
                case y5.Types.STRING:
                    return '<string>' + y5.Strings.escapeHTML(json) + '</string>';

                case y5.Types.NUMBER:
                    if (Math.round(json) == json) return '<int>' + y5.Strings.escapeHTML(json) + '</int>';
                    return '<double>' + y5.Strings.escapeHTML(json) + '</double>';

                case y5.Types.ARRAY:
                    var template = '<array><data>'
                    json.forEach(function(i){
                        template += '<value>';
                        template += json2value(i);
                        template += '</value>';
                    });
                    template += '</data></array>';
                    return template;

                case y5.Types.DATE:
                    // TODO: поддержать dateTime.iso8601
                    return '<string>' + y5.Strings.escapeHTML(json) + '</string>';

                case y5.Types.OBJECT:
                    var template = '<struct>'
                    for (var i in json) {
                        if (json.hasOwnProperty && !json.hasOwnProperty(i)) continue;
                        template += '<member>';
                        template += '<name>' + y5.Strings.escapeHTML(i) + '</name>';
                        template += '<value>';
                        template += json2value(json[i]);
                        template += '</value>';
                        template += '</member>';
                    }
                    template += '</struct>';
                    return template;

                default:
                    return null;
            }
        }
        var result = '<params>';
        json.forEach(function(i) {
            result += '<param><value>';
            result += json2value(i);
            result += '</value></param>';
        });
        result += '</params>';
        return result;
    },

    /**
     * Выполняет удалённый XML-RPC-вызов.
     * @memberOf y5.XmlRpc
     * @name call
     * @param {Object} params Объект параметров, поддерживаются:
     *   {String} method XML-RPC-метод
     *   {Array} params Параметры для XML-RPC, массив JavaScript-объектов
     *   {Function} onload Обработчик успешного вызова
     *   {Function} onerror Обработчик ошибки вызова, используется и при ошибках XML-RPC и при ошибках транспорта (y5.Request.XML)
     *
     * @example
     * xmlrpc.call({
     *     method: 'login',
     *     params: ['mylogin', 'mypassword']
     *     onload: function(res) { alert(res) },
     *     onerror: function(e) { alert(e) }
     * })
     */
    call: function(params) {
        var request = new y5.Request.XML(
            this.url,
            {
                method: 'post',
                contentType: 'text/xml'
            }
        );
        var _this = this;
        request.onload = function(req) {
            var result = _this.xmlrpc2json(req.responseXML);
            if ((y5.Types.type(result.param) != y5.Types.NULL && y5.Types.def(result.param)) && params.onload) {
                params.onload.call(this, result.param);
            };
            if (result.fault && params.onerror) {
                params.onerror.call(this, result.fault);
            };
        };
        request.onerror = function(e) {
            if (params.onerror) {
                params.onerror.call(this, e);
            }
        };
        request.send(
            '<?xml version="1.0"?>' +
            '<methodCall>' +
            '<methodName>' + params.method + '</methodName>' +
            _this.json2xmlrpc(params.params) +
            '</methodCall>'
        );
    }
};

y5.loaded('XmlRpc');
});
